# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import namedtuple
from typing import Any, ClassVar, Optional, Union

from graphql.type import GraphQLResolveInfo

from swh.graphql import resolvers as rs
from swh.graphql.backends.archive import Archive
from swh.graphql.errors import NullableObjectError, ObjectNotFoundError


class BaseNode:
    """
    Base resolver for all the nodes
    """

    _can_be_null: ClassVar[bool] = False

    def __init__(self, obj, info, node_data: Optional[Any] = None, **kwargs) -> None:
        self.obj: Optional[Union[BaseNode, rs.base_connection.BaseConnection]] = obj
        self.info: GraphQLResolveInfo = info
        self.kwargs = kwargs
        # initialize commonly used vars
        self.archive = Archive()
        self._node: Optional[Any] = self._get_node(node_data)
        # handle the errors, if any, after _node is set
        self._handle_node_errors()

    def _get_node(self, node_data: Optional[Any]) -> Optional[Any]:
        """
        Get the node object from the given data
        if the data (node_data) is none make a function call
        to get data from backend
        """
        if node_data is None:
            node_data = self._get_node_data()
        return self._get_node_from_data(node_data)

    def _get_node_from_data(self, node_data: Any) -> Optional[Any]:
        """
        Get the object from node_data
        In case of a dict, convert it to an object
        Override to support different data structures
        """
        if isinstance(node_data, dict):
            return namedtuple("NodeObj", node_data.keys())(*node_data.values())
        return node_data

    def _handle_node_errors(self) -> None:
        """
        Handle any error related to node data

        raise an error in case the object returned is None
        override for specific behaviour
        """
        if self._node is None and self._can_be_null:
            # fail silently
            raise NullableObjectError()
        elif self._node is None:
            # This will send this error to the client
            raise ObjectNotFoundError("Requested object is not available")

    def _get_node_data(self) -> Optional[Any]:
        """
        Override for desired behaviour
        This will be called only when node_data is None
        """
        # FIXME, make this call async (not for v1)

    def __getattr__(self, name: str) -> Any:
        """
        Any property defined in the sub-class will get precedence over
        the _node attributes
        """
        return getattr(self._node, name)

    def is_type_of(self) -> str:
        return self.__class__.__name__


class BaseSWHNode(BaseNode):
    """
    Base resolver for all the nodes with a SWHID field
    """

    @property
    def swhid(self):
        return self._node.swhid()



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Optional, Union

from swh.graphql.utils import utils
from swh.model.model import Revision, Sha1Git
from swh.model.swhids import CoreSWHID

from .base_connection import BaseConnection, ConnectionData
from .base_node import BaseSWHNode
from .target import BranchTargetNode, TargetNode


class BaseRevisionNode(BaseSWHNode):
    """
    Base resolver for all the revision nodes
    """

    def _get_revision_by_id(self, revision_id) -> Optional[Revision]:
        revisions = self.archive.get_revisions([revision_id])
        return revisions[0] if revisions else None

    @property
    def committerDate(self):  # To support the schema naming convention
        assert self._node is not None
        return self._node.committer_date

    @property
    def type(self) -> str:
        assert self._node is not None
        return self._node.type.value

    def is_type_of(self) -> str:
        # is_type_of is required only when resolving a UNION type
        # This is for ariadne to return the right type
        return "Revision"

    def directory_hash(self) -> Sha1Git:  # hash of the unique revision directory
        # To be used while resolving the revision directory
        assert self._node is not None
        return self._node.directory


class RevisionNode(BaseRevisionNode):
    """
    Node resolver for a revision requested directly with its SWHID
    """

    def _get_node_data(self) -> Optional[Revision]:
        revision_swhid = self.kwargs.get("swhid")
        assert isinstance(revision_swhid, CoreSWHID)
        return self._get_revision_by_id(revision_swhid.object_id)


class TargetRevisionNode(BaseRevisionNode):
    """
    Node resolver for a revision requested as a target
    """

    _can_be_null = True
    obj: Union[
        BranchTargetNode,
        TargetNode,
    ]

    def _get_node_data(self) -> Optional[Revision]:
        # self.obj.target_hash is the requested revision id
        return self._get_revision_by_id(self.obj.target_hash)


class ParentRevisionConnection(BaseConnection):
    """
    Connection resolver for parent revisions in a revision
    """

    obj: BaseRevisionNode

    _node_class = BaseRevisionNode

    def _get_connection_data(self) -> ConnectionData:
        # self.obj is the current(child) revision

        # FIXME, using dummy(local) pagination, move pagination to backend
        # STORAGE-TODO (pagination)
        parents = self.archive.get_revisions(self.obj.parents)
        return utils.get_local_paginated_data(
            parents, self._get_first_arg(), self._get_after_arg()
        )


class LogRevisionConnection(BaseConnection):
    """
    Connection resolver for the log (list of revisions) in a revision
    """

    obj: BaseRevisionNode

    _node_class = BaseRevisionNode

    def _get_connection_data(self) -> ConnectionData:
        log = self.archive.get_revision_log([self.obj.swhid.object_id])
        # Storage is returning a list of dicts instead of model objects
        # Following loop is to reverse that operation
        # STORAGE-TODO; remove to_dict from storage.revision_log
        log = [Revision.from_dict(rev) for rev in log]
        # FIXME, using dummy(local) pagination, move pagination to backend
        # STORAGE-TODO (pagination)
        return utils.get_local_paginated_data(
            log, self._get_first_arg(), self._get_after_arg()
        )



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import TYPE_CHECKING, Optional, Union

from swh.graphql.errors import NullableObjectError
from swh.graphql.utils import utils
from swh.model.model import CoreSWHID, Snapshot

from .base_connection import BaseConnection, ConnectionData
from .base_node import BaseSWHNode
from .origin import OriginNode
from .visit_status import BaseVisitStatusNode


class BaseSnapshotNode(BaseSWHNode):
    """
    Base resolver for all the snapshot nodes
    """

    def is_type_of(self):
        # is_type_of is required only when resolving a UNION type
        # This is for ariadne to return the right type
        return "Snapshot"


class SnapshotNode(BaseSnapshotNode):
    """
    Node resolver for a snapshot requested directly with its SWHID
    """

    def _get_node_data(self) -> Optional[Snapshot]:
        """ """
        swhid = self.kwargs.get("swhid")
        assert isinstance(swhid, CoreSWHID)
        return self.archive.get_snapshot(snapshot_id=swhid.object_id, verify=True)


class VisitSnapshotNode(BaseSnapshotNode):
    """
    Node resolver for a snapshot requested from a visit-status
    """

    # Visit snapshot is not using the target indirection,
    # hence this implementation (not using TargetSnapshotNode)

    _can_be_null = True
    obj: BaseVisitStatusNode

    def _get_node_data(self):
        snapshot_id = self.obj.snapshot_id()
        if snapshot_id is None:
            raise NullableObjectError()
        return self.archive.get_snapshot(snapshot_id=snapshot_id, verify=False)


class TargetSnapshotNode(BaseSnapshotNode):
    """
    Node resolver for a snapshot requested as a target
    """

    if TYPE_CHECKING:  # pragma: no cover
        from .target import BranchTargetNode

        obj: Union[BranchTargetNode]

    _can_be_null = True

    def _get_node_data(self):
        return self.archive.get_snapshot(snapshot_id=self.obj.target_hash, verify=False)


class LatestSnapshotNode(BaseSnapshotNode):
    """
    Node resolver for the latest snapshot in an origin
    """

    obj: OriginNode

    _can_be_null = True

    def _get_node_data(self):
        latest_status_with_snapshot = self.archive.get_latest_origin_visit_status(
            origin=self.obj.url,
            require_snapshot=True,
        )
        if not latest_status_with_snapshot:
            return None
        return self.archive.get_snapshot(
            snapshot_id=latest_status_with_snapshot.snapshot, verify=False
        )


class OriginSnapshotConnection(BaseConnection):
    """
    Connection resolver for the snapshots in an origin
    """

    obj: OriginNode

    _node_class = BaseSnapshotNode

    def _get_connection_data(self) -> ConnectionData:
        results = self.archive.get_origin_snapshots(self.obj.url)
        snapshots = [
            self.archive.get_snapshot(snapshot_id=snapshot, verify=False)
            for snapshot in results
        ]
        # FIXME, using dummy(local) pagination, move pagination to backend
        # To remove localpagination, just drop the paginated call
        # STORAGE-TODO
        return utils.get_local_paginated_data(
            snapshots, self._get_first_arg(), self._get_after_arg()
        )



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
High level resolvers
"""

# Any schema attribute can be resolved by any of the following ways
# and in the following priority order
# - In this module using a decorator (eg: @visit_status.field("snapshot"))
#   Every object (type) is expected to resolve this way as they can accept arguments
#   eg: origin.visits takes arguments to paginate
# - As a property in the Node object (eg: resolvers.visit.BaseVisitNode.id)
#   Every scalar is expected to resolve this way
# - As an attribute/item in the object/dict returned by a backend (eg: Origin.url)

import datetime
from typing import Optional, Union

from ariadne import ObjectType, UnionType
from graphql.type import GraphQLResolveInfo

from swh.graphql import resolvers as rs
from swh.graphql.utils import utils
from swh.model.model import TimestampWithTimezone

from .resolver_factory import (
    ConnectionObjectFactory,
    NodeObjectFactory,
    SimpleListFactory,
)

query: ObjectType = ObjectType("Query")
origin: ObjectType = ObjectType("Origin")
visit: ObjectType = ObjectType("Visit")
visit_status: ObjectType = ObjectType("VisitStatus")
snapshot: ObjectType = ObjectType("Snapshot")
snapshot_branch: ObjectType = ObjectType("Branch")
revision: ObjectType = ObjectType("Revision")
release: ObjectType = ObjectType("Release")
directory: ObjectType = ObjectType("Directory")
directory_entry: ObjectType = ObjectType("DirectoryEntry")
binary_string: ObjectType = ObjectType("BinaryString")
date: ObjectType = ObjectType("Date")
branch_target: ObjectType = ObjectType("BranchTarget")
release_target: ObjectType = ObjectType("ReleaseTarget")
directory_entry_target: ObjectType = ObjectType("DirectoryEntryTarget")
origin_search_result: ObjectType = ObjectType("OriginSearchResult")

branch_target_node: UnionType = UnionType("BranchTargetNode")
release_target_node: UnionType = UnionType("ReleaseTargetNode")
directory_entry_target_node: UnionType = UnionType("DirectoryEntryTargetNode")
resolve_swhid_result: UnionType = UnionType("ResolveSWHIDResult")


# Node resolvers
# A node resolver will return either an instance of a BaseNode subclass or None


@query.field("origin")
def origin_resolver(obj: None, info: GraphQLResolveInfo, **kw) -> rs.origin.OriginNode:
    return NodeObjectFactory.create("origin", obj, info, **kw)


@origin.field("latestVisit")
def latest_visit_resolver(
    obj: rs.origin.BaseOriginNode, info: GraphQLResolveInfo, **kw
) -> Optional[rs.visit.LatestVisitNode]:
    return NodeObjectFactory.create("latest-visit", obj, info, **kw)


@origin.field("latestSnapshot")
def latest_snapshot_resolver(
    obj: rs.origin.BaseOriginNode, info: GraphQLResolveInfo, **kw
) -> Optional[rs.snapshot.LatestSnapshotNode]:
    return NodeObjectFactory.create("latest-snapshot", obj, info, **kw)


@query.field("visit")
def visit_resolver(
    obj: None, info: GraphQLResolveInfo, **kw
) -> rs.visit.OriginVisitNode:
    return NodeObjectFactory.create("visit", obj, info, **kw)


@visit.field("latestStatus")
def latest_visit_status_resolver(
    obj: rs.visit.BaseVisitNode, info: GraphQLResolveInfo, **kw
) -> Optional[rs.visit_status.LatestVisitStatusNode]:
    return NodeObjectFactory.create("latest-status", obj, info, **kw)


@query.field("snapshot")
def snapshot_resolver(
    obj: None, info: GraphQLResolveInfo, **kw
) -> rs.snapshot.SnapshotNode:
    return NodeObjectFactory.create("snapshot", obj, info, **kw)


@visit_status.field("snapshot")
def visit_snapshot_resolver(
    obj: rs.visit_status.BaseVisitStatusNode, info: GraphQLResolveInfo, **kw
) -> Optional[rs.snapshot.VisitSnapshotNode]:
    return NodeObjectFactory.create("visit-snapshot", obj, info, **kw)


@query.field("revision")
def revision_resolver(
    obj: None, info: GraphQLResolveInfo, **kw
) -> rs.revision.RevisionNode:
    return NodeObjectFactory.create("revision", obj, info, **kw)


@revision.field("directory")
def revision_directory_resolver(
    obj: rs.revision.BaseRevisionNode, info: GraphQLResolveInfo, **kw
) -> Optional[rs.directory.RevisionDirectoryNode]:
    return NodeObjectFactory.create("revision-directory", obj, info, **kw)


@query.field("release")
def release_resolver(
    obj: None, info: GraphQLResolveInfo, **kw
) -> rs.release.ReleaseNode:
    return NodeObjectFactory.create("release", obj, info, **kw)


@query.field("directory")
def directory_resolver(
    obj: None, info: GraphQLResolveInfo, **kw
) -> rs.directory.DirectoryNode:
    return NodeObjectFactory.create("directory", obj, info, **kw)


@query.field("directoryEntry")
def directory_entry_resolver(
    obj: None, info: GraphQLResolveInfo, **kw
) -> rs.directory_entry.DirectoryEntryNode:
    return NodeObjectFactory.create("directory-entry", obj, info, **kw)


@directory_entry.field("target")
@release.field("target")
def generic_target_resolver(
    obj: Union[rs.release.BaseReleaseNode, rs.directory_entry.BaseDirectoryEntryNode],
    info: GraphQLResolveInfo,
    **kw,
) -> rs.target.TargetNode:
    return NodeObjectFactory.create("generic-target", obj, info, **kw)


@snapshot_branch.field("target")
def snapshot_branch_target_resolver(
    obj: rs.snapshot_branch.SnapshotBranchNode, info: GraphQLResolveInfo, **kw
) -> rs.target.BranchTargetNode:
    # Snapshot branch target is a special case
    return NodeObjectFactory.create("branch-target", obj, info, **kw)


@directory_entry_target.field("node")
@release_target.field("node")
@branch_target.field("node")
def generic_target_node_resolver(
    obj: rs.target.TargetNode, info: GraphQLResolveInfo, **kw
) -> Optional[
    Union[
        rs.revision.BaseRevisionNode,
        rs.release.BaseReleaseNode,
        rs.directory.BaseDirectoryNode,
        rs.content.BaseContentNode,
        rs.snapshot.BaseSnapshotNode,
    ]
]:
    if not obj or not obj.type:
        # Target can be None for a branch, return None in that case
        return None
    return NodeObjectFactory.create(f"target-{obj.type}", obj, info, **kw)


@query.field("contentByHashes")
def content_by_hashes_resolver(
    obj: None, info: GraphQLResolveInfo, **kw
) -> rs.content.ContentbyHashesNode:
    return NodeObjectFactory.create("content-by-hashes", obj, info, **kw)


@origin_search_result.field("node")
def origin_search_node_resolver(
    obj: None, info: GraphQLResolveInfo, **kw
) -> rs.origin.TargetOriginNode:
    return NodeObjectFactory.create("target-origin", obj, info, **kw)


# Connection resolvers
# A connection resolver should return an instance of BaseConnection


@query.field("origins")
def origins_resolver(
    obj: None, info: GraphQLResolveInfo, **kw
) -> rs.origin.OriginConnection:
    return ConnectionObjectFactory.create("origins", obj, info, **kw)


@origin.field("visits")
def visits_resolver(
    obj: rs.origin.BaseOriginNode, info: GraphQLResolveInfo, **kw
) -> rs.visit.OriginVisitConnection:
    return ConnectionObjectFactory.create("origin-visits", obj, info, **kw)


@origin.field("snapshots")
def origin_snapshots_resolver(
    obj: rs.origin.BaseOriginNode, info: GraphQLResolveInfo, **kw
) -> rs.snapshot.OriginSnapshotConnection:
    return ConnectionObjectFactory.create("origin-snapshots", obj, info, **kw)


@visit.field("statuses")
def visitstatus_resolver(
    obj: rs.visit.BaseVisitNode, info: GraphQLResolveInfo, **kw
) -> rs.visit_status.VisitStatusConnection:
    return ConnectionObjectFactory.create("visit-status", obj, info, **kw)


@snapshot.field("branches")
def snapshot_branches_resolver(
    obj: rs.snapshot.BaseSnapshotNode, info: GraphQLResolveInfo, **kw
) -> rs.snapshot_branch.SnapshotBranchConnection:
    return ConnectionObjectFactory.create("snapshot-branches", obj, info, **kw)


@revision.field("parents")
def revision_parents_resolver(
    obj: rs.revision.BaseRevisionNode, info: GraphQLResolveInfo, **kw
) -> rs.revision.ParentRevisionConnection:
    return ConnectionObjectFactory.create("revision-parents", obj, info, **kw)


@revision.field("revisionLog")
def revision_log_resolver(
    obj: rs.revision.BaseRevisionNode, info: GraphQLResolveInfo, **kw
) -> rs.revision.LogRevisionConnection:
    return ConnectionObjectFactory.create("revision-log", obj, info, **kw)


@directory.field("entries")
def directory_entries_resolver(
    obj: rs.directory.BaseDirectoryNode, info: GraphQLResolveInfo, **kw
) -> rs.directory_entry.DirectoryEntryConnection:
    return ConnectionObjectFactory.create("directory-entries", obj, info, **kw)


@query.field("originSearch")
def origin_search_resolver(
    obj: None, info: GraphQLResolveInfo, **kw
) -> rs.search.OriginSearchConnection:
    return ConnectionObjectFactory.create("origin-search", obj, info, **kw)


# Simple list resolvers (lists without paginations)


@query.field("contentsBySWHID")
def contnets_by_swhid_resolver(
    obj: None, info: GraphQLResolveInfo, **kw
) -> rs.content.ContentSwhidList:
    return SimpleListFactory.create("contents-swhid", obj, info, **kw)


@query.field("contentsByHashes")
def contnets_by_hashes_resolver(
    obj: None, info: GraphQLResolveInfo, **kw
) -> rs.content.ContentHashList:
    return SimpleListFactory.create("contents-hashes", obj, info, **kw)


@query.field("resolveSWHID")
def resolve_swhid_resolver(
    obj: None, info: GraphQLResolveInfo, **kw
) -> rs.swhid.ResolveSWHIDList:
    return SimpleListFactory.create("resolve-swhid", obj, info, **kw)


@revision.field("author")
def revision_author_resolver(
    obj: None, info: GraphQLResolveInfo, **kw
) -> rs.revision.RevisionNode:
    return SimpleListFactory.create("revision-author", obj, info, **kw)


@revision.field("committer")
def revision_committer_resolver(
    obj: None, info: GraphQLResolveInfo, **kw
) -> rs.revision.RevisionNode:
    return SimpleListFactory.create("revision-committer", obj, info, **kw)


@release.field("author")
def release_author_resolver(
    obj: rs.release.BaseReleaseNode, info: GraphQLResolveInfo, **kw
) -> rs.revision.RevisionNode:
    return SimpleListFactory.create("release-author", obj, info, **kw)


# Other resolvers


@release_target_node.type_resolver
@directory_entry_target_node.type_resolver
@branch_target_node.type_resolver
@resolve_swhid_result.type_resolver
def union_resolver(
    obj: Union[
        rs.revision.BaseRevisionNode,
        rs.release.BaseReleaseNode,
        rs.directory.BaseDirectoryNode,
        rs.content.BaseContentNode,
        rs.snapshot.BaseSnapshotNode,
    ],
    *_,
) -> str:
    """
    Generic resolver for all the union types
    """
    return obj.is_type_of()


# BinaryString resolvers


@binary_string.field("text")
def binary_string_text_resolver(obj: bytes, *args, **kw) -> str:
    return obj.decode(utils.ENCODING, "replace")


@binary_string.field("base64")
def binary_string_base64_resolver(obj: bytes, *args, **kw) -> str:
    return utils.get_b64_string(obj)


# Date object resolver


@date.field("date")
def date_date_resolver(
    obj: TimestampWithTimezone, *args: GraphQLResolveInfo, **kw
) -> datetime.datetime:
    # This will be serialised as a DateTime Scalar
    return obj.to_datetime()


@date.field("offset")
def date_offset_resolver(
    obj: TimestampWithTimezone, *args: GraphQLResolveInfo, **kw
) -> bytes:
    # This will be serialised as a Binary string
    return obj.offset_bytes



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from . import utils
from ..data import get_origins, get_visit_status, get_visits


@pytest.mark.parametrize(
    "visit, visit_status", list(zip(get_visits(), get_visit_status()))
)
def test_get_visit_status(client, visit, visit_status):
    query_str = """
    query getVisit($origin: String!, $visitId: Int!) {
      visit(originUrl: $origin, visitId: $visitId) {
        statuses(first: 3) {
          nodes {
            status
            date
            type
            snapshot {
              swhid
            }
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(
        client, query_str, origin=visit.origin, visitId=visit.visit
    )
    assert data["visit"]["statuses"]["nodes"][0] == {
        "date": visit_status.date.isoformat(),
        "snapshot": {"swhid": f"swh:1:snp:{visit_status.snapshot.hex()}"}
        if visit_status.snapshot is not None
        else None,
        "status": visit_status.status,
        "type": visit_status.type,
    }


def test_visit_status_pagination(client):
    # visit status is using a different cursor, hence separate test
    query_str = """
    query getVisit($origin: String!, $visitId: Int!) {
      visit(originUrl: $origin, visitId: $visitId) {
        statuses(first: 1) {
          pageInfo {
            hasNextPage
            endCursor
          }
          edges {
            node {
              status
            }
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(
        client, query_str, origin=get_origins()[0].url, visitId=1
    )
    # request again with the endcursor
    end_cursor = data["visit"]["statuses"]["pageInfo"]["endCursor"]
    query_str = """
    query getVisit($origin: String!, $visitId: Int!, $after: String) {
      visit(originUrl: $origin, visitId: $visitId) {
        statuses(first: 1, after: $after) {
          pageInfo {
            hasNextPage
            endCursor
          }
          edges {
            node {
              status
            }
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(
        client,
        query_str,
        origin=get_origins()[0].url,
        visitId=1,
        after=end_cursor,
    )
    assert data["visit"]["statuses"] == {
        "edges": [
            {
                "node": {"status": "ongoing"},
            }
        ],
        "pageInfo": {"endCursor": None, "hasNextPage": False},
    }



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.graphql.resolvers import resolver_factory


class TestFactory:
    def test_get_node_resolver_invalid_type(self):
        with pytest.raises(AttributeError):
            resolver_factory.NodeObjectFactory().create("invalid", None, None)

    def test_get_connection_resolver_invalid_type(self):
        with pytest.raises(AttributeError):
            resolver_factory.ConnectionObjectFactory().create("invalid", None, None)

    def test_get_base_list_resolver_invalid_type(self):
        with pytest.raises(AttributeError):
            resolver_factory.SimpleListFactory().create("invalid", None, None)



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import List, Optional

from swh.model.model import Person as StoragePerson

from .base_connection import BaseList
from .base_node import BaseNode
from .release import BaseReleaseNode
from .revision import BaseRevisionNode


class Person(BaseNode):
    pass


def get_person_list(person: Optional[StoragePerson]) -> List[StoragePerson]:
    return [person] if person else []


class RevisionAuthorList(BaseList):
    """
    List of revision authors
    """

    obj: BaseRevisionNode

    _node_class = Person

    def _get_results(self) -> List[StoragePerson]:
        """
        Author is a single object in the current data model,
        return it as a list to support future evolutions

        No backend fetching required as the data exists in
        the parent object (revision)
        """

        return get_person_list(self.obj.author)


class RevisionCommitterList(BaseList):
    obj: BaseRevisionNode

    _node_class = Person

    def _get_results(self) -> List[StoragePerson]:
        """
        Committer is a single object in the current data model,
        return it as a list to support future evolutions

        No backend fetching required as the data exists in
        the parent object (revision)
        """

        return get_person_list(self.obj.committer)


class ReleaseAuthorList(BaseList):
    obj: BaseReleaseNode

    _node_class = Person

    def _get_results(self) -> List[StoragePerson]:
        """
        Author is a single object in the current data model,
        return it as a list to support future evolutions

        No backend fetching required as the data exists in
        the parent object (release)
        """

        return get_person_list(self.obj.author)



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import List, Optional, Tuple

from swh.graphql.errors import DataError
from swh.graphql.utils import utils
from swh.model.model import CoreSWHID, SnapshotBranch
from swh.storage.interface import PagedResult

from .base_connection import BaseConnection, ConnectionData
from .base_node import BaseNode


class SnapshotBranchNode(BaseNode):

    obj: "SnapshotBranchConnection"

    def _get_node_from_data(self, node_data: Tuple[bytes, Optional[SnapshotBranch]]):
        # node_data is a tuple as returned by _get_connection_data in SnapshotBranchConnection
        # overriding to support this special data structure
        branch_name, branch_obj = node_data
        updated_node_data = {
            # Name of the branch, exposed in the schema
            "name": branch_name,
            # Type of the branch, exposed in the schema
            "type": branch_obj.target_type.value if branch_obj else None,
            # not exposed in the schema, to be used by the target object
            "snapshot_id": self._get_snapshot_swhid().object_id,
            # not exposed in the schema, to be used by the target object
            "branch_obj": branch_obj,
        }
        return super()._get_node_from_data(updated_node_data)

    def _get_snapshot_swhid(self) -> CoreSWHID:
        from .snapshot import BaseSnapshotNode

        # As of now parent of SnapshotBranch will always be a SnapshotBranchConnection
        # and self.obj.obj will always be a BaseSnapshot object
        # override this method for SnapshotBranch nodes reachable by some other path
        if not isinstance(self.obj.obj, BaseSnapshotNode):
            # This is not expected to happen with the existing endpoints
            raise DataError(
                "SnapshotBranch is accessed outside a SnapshotBranchConnection context"
            )
        return self.obj.obj.swhid


class SnapshotBranchConnection(BaseConnection):
    """
    Connection resolver for the branches in a snapshot
    """

    from .snapshot import BaseSnapshotNode

    obj: BaseSnapshotNode

    _node_class = SnapshotBranchNode

    def _get_connection_data(self) -> ConnectionData:
        branches = self.archive.get_snapshot_branches(
            snapshot=self.obj.swhid.object_id,
            after=self._get_after_arg(),
            first=self._get_first_arg(),
            target_types=self.kwargs.get("types"),
            name_include=self._get_name_include_arg(),
            name_exclude_prefix=self._get_name_exclude_prefix_arg(),
        )
        end_cursor: Optional[bytes] = branches.get("next_branch") if branches else None
        # FIXME, this pagination is not consistent with other connections
        # FIX in swh-storage to return PagedResult
        # STORAGE-TODO

        # each result item will be converted to a dict in _get_node_from_data
        # method in the node class
        results: List[Tuple[bytes, Optional[SnapshotBranch]]] = (
            list(branches["branches"].items()) if branches else []
        )
        return ConnectionData(
            paged_result=PagedResult(
                results=results,
                next_page_token=end_cursor.decode() if end_cursor else None,
            )
        )

    def _get_after_arg(self):
        # after argument must be an empty string by default
        after = super()._get_after_arg()
        return after.encode() if after else b""

    def _get_name_include_arg(self):
        name_include = self.kwargs.get("nameInclude", None)
        return name_include.encode() if name_include else None

    def _get_name_exclude_prefix_arg(self):
        name_exclude_prefix = self.kwargs.get("nameExcludePrefix", None)
        return name_exclude_prefix.encode() if name_exclude_prefix else None

    def _get_index_cursor(self, index: int, node: SnapshotBranchNode):
        # Snapshot branch is using a different cursor, hence the override
        return utils.get_encoded_cursor(node.name)



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from . import utils
from ..data import get_origins


@pytest.mark.parametrize("origin", get_origins())
def test_get_visit(client, storage, origin):
    query_str = """
    query getVisit($origin: String!, $visitId: Int!) {
      visit(originUrl: $origin, visitId: $visitId) {
        visitId
        date
        type
        latestStatus {
          status
          date
          type
          snapshot {
            swhid
          }
        }
        statuses {
          nodes {
            status
          }
        }
      }
    }
    """

    visits_and_statuses = storage.origin_visit_get_with_statuses(origin.url).results
    for vws in visits_and_statuses:
        visit = vws.visit
        statuses = vws.statuses
        data, _ = utils.get_query_response(
            client, query_str, origin=origin.url, visitId=visit.visit
        )
        assert data["visit"] == {
            "visitId": visit.visit,
            "type": visit.type,
            "date": visit.date.isoformat(),
            "latestStatus": {
                "date": statuses[-1].date.isoformat(),
                "type": statuses[-1].type,
                "status": statuses[-1].status,
                "snapshot": ({"swhid": f"swh:1:snp:{statuses[-1].snapshot.hex()}"})
                if statuses[-1].snapshot
                else None,
            },
            "statuses": {"nodes": [{"status": status.status} for status in statuses]},
        }


def test_invalid_get_visit(client):
    query_str = """
    {
      visit(originUrl: "http://example.com/forge1", visitId: 3) {
        type
      }
    }
    """
    utils.assert_missing_object(client, query_str, "visit")


@pytest.mark.parametrize("sort", ["ASC", "DESC"])
def test_visit_status_sort_order(client, storage, sort):
    query_str = """
    query getVisit($origin: String!, $visitId: Int!, $sort: ListOrder) {
      visit(originUrl: $origin, visitId: $visitId) {
        visitId
        statuses(sort: $sort) {
          nodes {
            date
          }
        }
      }
    }
    """
    # Testing the only test visit object with multiple statuses
    origin = get_origins()[0].url
    visit = 1
    response, _ = utils.get_query_response(
        client, query_str, origin=origin, visitId=visit, sort=sort
    )
    data_statuses = response["visit"]["statuses"]["nodes"]
    if sort == "DESC":
        data_statuses.reverse()
    original_data = [
        {"date": x.date.isoformat()}
        for x in storage.origin_visit_status_get(origin=origin, visit=visit).results
    ]
    assert data_statuses == original_data


def test_get_latest_visit_status_filter_by_status_return_null(client):
    query_str = """
    query getVisit($origin: String!, $visitId: Int!) {
      visit(originUrl: $origin, visitId: $visitId) {
        visitId
        date
        type
        latestStatus(allowedStatuses: [full]) {
          status
        }
      }
    }
    """
    data, err = utils.get_query_response(
        client, query_str, origin=get_origins()[0].url, visitId=1
    )
    assert err is None
    assert data == {
        "visit": {
            "date": "2013-05-07T04:20:39.369271+00:00",
            "latestStatus": None,
            "type": "git",
            "visitId": 1,
        }
    }


def test_get_latest_visit_status_filter_by_type(client):
    query_str = """
    query getVisit($origin: String!, $visitId: Int!) {
      visit(originUrl: $origin, visitId: $visitId) {
        visitId
        date
        type
        latestStatus(allowedStatuses: [ongoing]) {
          status
          date
        }
      }
    }
    """
    data, err = utils.get_query_response(
        client, query_str, origin=get_origins()[0].url, visitId=1
    )
    assert err is None
    assert data == {
        "visit": {
            "date": "2013-05-07T04:20:39.369271+00:00",
            "latestStatus": {
                "date": "2014-05-07T04:20:39.432222+00:00",
                "status": "ongoing",
            },
            "type": "git",
            "visitId": 1,
        }
    }


def test_get_latest_visit_status_filter_by_snapshot(client):
    query_str = """
    query getVisit($origin: String!, $visitId: Int!) {
      visit(originUrl: $origin, visitId: $visitId) {
        visitId
        date
        type
        latestStatus(requireSnapshot: true) {
          status
          date
          snapshot {
            swhid
          }
        }
      }
    }
    """
    data, err = utils.get_query_response(
        client, query_str, origin=get_origins()[1].url, visitId=2
    )
    assert err is None
    assert data == {
        "visit": {
            "date": "2015-11-27T17:20:39+00:00",
            "latestStatus": {
                "date": "2015-11-27T17:22:18+00:00",
                "snapshot": {
                    "swhid": "swh:1:snp:0e7f84ede9a254f2cd55649ad5240783f557e65f"
                },
                "status": "partial",
            },
            "type": "hg",
            "visitId": 2,
        }
    }



"""
# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information
"""

from . import utils


def test_auth_error(client, authentication_error):
    query_str = """
    query getOrigin($url: String!) {
      origin(url: $url) {
        url
      }
    }
    """
    errors = utils.get_error_response(
        client, query_str, response_code=401, url="http://example.com"
    )
    assert "Authentication error" in errors[0]["message"]



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from typing import ClassVar, Dict, Type

from swh.core import statsd
from swh.graphql.errors import NullableObjectError

from .base_connection import BaseConnection, BaseList
from .base_node import BaseNode
from .content import (
    ContentbyHashesNode,
    ContentHashList,
    ContentSwhidList,
    TargetContentNode,
)
from .directory import DirectoryNode, RevisionDirectoryNode, TargetDirectoryNode
from .directory_entry import DirectoryEntryConnection, DirectoryEntryNode
from .origin import OriginConnection, OriginNode, TargetOriginNode
from .person import ReleaseAuthorList, RevisionAuthorList, RevisionCommitterList
from .release import ReleaseNode, TargetReleaseNode
from .revision import (
    LogRevisionConnection,
    ParentRevisionConnection,
    RevisionNode,
    TargetRevisionNode,
)
from .search import OriginSearchConnection
from .snapshot import (
    LatestSnapshotNode,
    OriginSnapshotConnection,
    SnapshotNode,
    TargetSnapshotNode,
    VisitSnapshotNode,
)
from .snapshot_branch import SnapshotBranchConnection
from .swhid import ResolveSWHIDList
from .target import BranchTargetNode, TargetNode
from .visit import LatestVisitNode, OriginVisitConnection, OriginVisitNode
from .visit_status import LatestVisitStatusNode, VisitStatusConnection

this_statsd = statsd.Statsd(namespace="swh_graphql")
logger = logging.getLogger(__name__)


class NodeObjectFactory:
    mapping: ClassVar[Dict[str, Type[BaseNode]]] = {
        "origin": OriginNode,
        "visit": OriginVisitNode,
        "latest-visit": LatestVisitNode,
        "latest-status": LatestVisitStatusNode,
        "latest-snapshot": LatestSnapshotNode,
        "visit-snapshot": VisitSnapshotNode,
        "snapshot": SnapshotNode,
        "revision": RevisionNode,
        "revision-directory": RevisionDirectoryNode,
        "release": ReleaseNode,
        "directory": DirectoryNode,
        "directory-entry": DirectoryEntryNode,
        "content-by-hashes": ContentbyHashesNode,
        "generic-target": TargetNode,
        "branch-target": BranchTargetNode,
        "target-origin": TargetOriginNode,
        "target-snapshot": TargetSnapshotNode,
        "target-revision": TargetRevisionNode,
        "target-release": TargetReleaseNode,
        "target-directory": TargetDirectoryNode,
        "target-content": TargetContentNode,
    }

    @classmethod
    def create(cls, node_type: str, obj, info, *args, **kw):
        # FIXME, add to the sentry transaction
        resolver = cls.mapping.get(node_type)
        if not resolver:
            raise AttributeError(f"Invalid node type: {node_type}")
        with this_statsd.timed("node_query_seconds", tags={"node": node_type}):
            try:
                node_obj = resolver(obj, info, *args, **kw)
            except NullableObjectError:
                # Return None instead of the object
                # FIXME, add to the sentry transaction
                logger.warning("Null %s object", node_type)
                node_obj = None
        return node_obj


class ConnectionObjectFactory:
    mapping: ClassVar[Dict[str, Type[BaseConnection]]] = {
        "origins": OriginConnection,
        "origin-visits": OriginVisitConnection,
        "origin-snapshots": OriginSnapshotConnection,
        "visit-status": VisitStatusConnection,
        "snapshot-branches": SnapshotBranchConnection,
        "revision-parents": ParentRevisionConnection,
        "revision-log": LogRevisionConnection,
        "directory-entries": DirectoryEntryConnection,
        "origin-search": OriginSearchConnection,
    }

    @classmethod
    def create(cls, connection_type: str, obj, info, *args, **kw):
        # FIXME, add to the sentry transaction
        resolver = cls.mapping.get(connection_type)
        if not resolver:
            raise AttributeError(f"Invalid connection type: {connection_type}")
        with this_statsd.timed(
            "connection_query_seconds", tags={"connection": connection_type}
        ):
            return resolver(obj, info, *args, **kw)


class SimpleListFactory:
    mapping: ClassVar[Dict[str, Type[BaseList]]] = {
        "resolve-swhid": ResolveSWHIDList,
        "revision-author": RevisionAuthorList,
        "revision-committer": RevisionCommitterList,
        "release-author": ReleaseAuthorList,
        "contents-swhid": ContentSwhidList,
        "contents-hashes": ContentHashList,
    }

    @classmethod
    def create(cls, list_type: str, obj, info, *args, **kw):
        # FIXME, add to the sentry transaction
        resolver = cls.mapping.get(list_type)
        if not resolver:
            raise AttributeError(f"Invalid list type: {list_type}")
        with this_statsd.timed("list_query_seconds", tags={"list": list_type}):
            # invoke the get_results method to return the list
            return resolver(obj, info, *args, **kw).get_results()



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import List, Optional, Union

from swh.graphql.errors import DataError, InvalidInputError
from swh.model import hashutil
from swh.model.model import Content
from swh.storage.interface import HashDict, TotalHashDict

from .base_connection import BaseList
from .base_node import BaseSWHNode
from .target import BranchTargetNode, TargetNode


def read_and_validate_content_hashes(hashes):
    try:
        return {
            hash_type: hashutil.hash_to_bytes(hash_value)
            for (hash_type, hash_value) in hashes
        }
    except ValueError as e:
        # raise an input error in case of an invalid hash
        raise InvalidInputError("Invalid content hash", e)


class BaseContentNode(BaseSWHNode):
    """
    Base resolver for all the content nodes
    """

    @property
    def hashes(self):
        # FIXME, use a Node instead
        return {k: v.hex() for (k, v) in self._node.hashes().items()}

    @property
    def id(self):
        return self._node.sha1_git

    @property
    def data(self):
        # FIXME, return a Node object
        # FIXME, add more ways to retrieve data like binary string
        archive_url = "https://archive.softwareheritage.org/api/1/"
        content_sha1 = self._node.hashes()["sha1"]
        return {
            "url": f"{archive_url}content/sha1:{content_sha1.hex()}/raw/",
        }

    @property
    def mimeType(self):
        # FIXME, fetch data from the indexers
        return None

    @property
    def language(self):
        # FIXME, fetch data from the indexers
        return None

    @property
    def license(self):
        # FIXME, fetch data from the indexers
        return None

    def is_type_of(self):
        # is_type_of is required only when resolving a UNION type
        # This is for ariadne to return the right type
        return "Content"


class ContentbyHashesNode(BaseContentNode):
    """
    Node resolver for a content requested with all of its hashes
    A single content object will be returned
    """

    def _get_node_data(self) -> Optional[Content]:
        hashes: TotalHashDict = read_and_validate_content_hashes(self.kwargs.items())
        contents = self.archive.get_contents(hashes=hashes)
        if len(contents) > 1:
            # This situation is not expected to happen IRL
            raise DataError("Content hash conflict for the set ", hashes)
        return contents[0] if contents else None


class TargetContentNode(BaseContentNode):
    """
    Node resolver for a content requested as a target
    """

    _can_be_null = True
    obj: Union[
        TargetNode,
        BranchTargetNode,
    ]

    def _get_node_data(self) -> Optional[Content]:
        # FIXME, this is not considering hash collisions
        # and could return a wrong object in very rare situations
        contents = self.archive.get_contents(hashes={"sha1_git": self.obj.target_hash})
        # always returning the first content from the storage
        return contents[0] if contents else None


class ContentSwhidList(BaseList):
    """
    Return a non paginated list of contents for the given SWHID
    This will return a single item in most of the cases
    """

    _node_class = BaseContentNode

    def _get_results(self) -> List[Content]:
        hashes = {"sha1_git": self.kwargs.get("swhid").object_id}
        return self.archive.get_contents(hashes=hashes)


class ContentHashList(BaseList):
    """
    Return a non paginated list of contents for the given hashes
    This will return a single item in most of the cases
    """

    _node_class = BaseContentNode

    def _get_results(self) -> List[Content]:
        hashes: HashDict = read_and_validate_content_hashes(self.kwargs.items())
        if not hashes:
            raise InvalidInputError("At least one of the four hashes must be provided")
        return self.archive.get_contents(hashes=hashes)



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from typing import Any, Optional

import pytest
from starlette.authentication import AuthCredentials, AuthenticationError, SimpleUser
from starlette.testclient import TestClient

from swh.auth.starlette.backends import BearerTokenAuthBackend
from swh.graphql import server as app_server

from .data import populate_dummy_data, populate_search_data


def mock_async(return_value: Any = None, error: Optional[Exception] = None):
    # FIXME, mock function for async. Can be removed on Python.3.8+
    async def func(*args, **kw):
        if error:
            raise error
        return return_value

    return func


@pytest.fixture(scope="session")
def test_app():
    os.environ["SWH_CONFIG_FILENAME"] = os.path.join(
        os.path.dirname(__file__), "../config/test.yml"
    )
    app = app_server.make_app_from_configfile()
    return app


@pytest.fixture(autouse=True)
def authenticated_user(mocker):
    return_value = AuthCredentials(["Authenticated"]), SimpleUser("user")
    mocker.patch.object(
        BearerTokenAuthBackend,
        "authenticate",
        side_effect=mock_async(return_value=return_value),
    )


@pytest.fixture
def anonymous_user(mocker):
    mocker.patch.object(
        BearerTokenAuthBackend,
        "authenticate",
        side_effect=mock_async(return_value=None),
    )


@pytest.fixture
def authentication_error(mocker):
    mocker.patch.object(
        BearerTokenAuthBackend,
        "authenticate",
        side_effect=mock_async(error=AuthenticationError("auth error")),
    )


@pytest.fixture(scope="session")
def storage(test_app):
    storage = app_server.get_storage()
    # populate the in-memory storage
    populate_dummy_data(storage)
    return storage


@pytest.fixture(scope="session")
def search(test_app):
    search = app_server.get_search()
    search.initialize()
    # populate the in-memory search
    populate_search_data(search)
    return search


@pytest.fixture(scope="session")
def client(test_app, storage, search):
    yield TestClient(test_app)


@pytest.fixture
def high_query_cost():
    query_cost = app_server.graphql_cfg["max_query_cost"]["user"]
    app_server.graphql_cfg["max_query_cost"]["user"] = 2000
    yield
    app_server.graphql_cfg["max_query_cost"]["user"] = query_cost


@pytest.fixture
def none_query_cost():
    # There will not be any query cost limit in this case
    query_cost = app_server.graphql_cfg["max_query_cost"]["user"]
    app_server.graphql_cfg["max_query_cost"]["user"] = None
    yield
    app_server.graphql_cfg["max_query_cost"]["user"] = query_cost



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
from typing import Optional

from ariadne import ScalarType

from swh.graphql.errors import InvalidInputError
from swh.graphql.utils import utils
from swh.model.exceptions import ValidationError
from swh.model.swhids import CoreSWHID

datetime_scalar = ScalarType("DateTime")
swhid_scalar = ScalarType("SWHID")
id_scalar = ScalarType("ID")


@id_scalar.serializer
def serialize_id(value) -> str:
    if isinstance(value, str):
        value = value.encode()
    return value.hex()


@datetime_scalar.serializer
def serialize_datetime(value: Optional[datetime.datetime]) -> Optional[str]:
    return None if value is None else utils.get_formatted_date(value)


@swhid_scalar.value_parser
def validate_swhid(value):
    try:
        swhid = CoreSWHID.from_string(value)
    except ValidationError as e:
        raise InvalidInputError("Invalid SWHID", e)
    return swhid


@swhid_scalar.serializer
def serialize_swhid(value):
    return str(value)



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from . import utils


def test_search_origins(client):
    query_str = """
    query doSearch($query: String!, $first: Int!) {
      originSearch(query: $query, first: $first) {
        nodes {
          type
          url
          node {
            url
            latestVisit {
              date
            }
          }
        }
        pageInfo {
          hasNextPage
          endCursor
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, query="fox", first=1)
    assert len(data["originSearch"]["nodes"]) == 1
    assert data == {
        "originSearch": {
            "nodes": [
                {
                    "type": "origin",
                    "url": "https://somewhere.org/den/fox",
                    "node": {
                        "url": "https://somewhere.org/den/fox",
                        "latestVisit": {"date": "2018-11-27T17:20:39+00:00"},
                    },
                }
            ],
            "pageInfo": {"endCursor": "MQ==", "hasNextPage": True},
        }
    }


def test_search_missing_url(client):
    query_str = """
    query doSearch($query: String!, $first: Int!) {
      originSearch(query: $query, first: $first) {
        nodes {
          type
        }
        pageInfo {
          hasNextPage
          endCursor
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, query="missing-fox", first=1)
    assert len(data["originSearch"]["nodes"]) == 0



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.model.model import Origin
from swh.storage.interface import PagedResult

from .base_connection import BaseConnection, ConnectionData
from .base_node import BaseSWHNode
from .search import OriginSearchResultNode


class BaseOriginNode(BaseSWHNode):
    pass


class OriginNode(BaseOriginNode):
    """
    Node resolver for an origin requested directly with its URL
    """

    def _get_node_data(self):
        return self.archive.get_origin(self.kwargs.get("url"))


class TargetOriginNode(BaseOriginNode):
    """
    Node resolver for an origin requested as a target
    """

    obj: OriginSearchResultNode
    _can_be_null = True

    def _get_node_data(self):
        # The target origin URL is guaranteed to exist in the archive
        # Hence returning the origin object without any explicit check in the archive
        # This assumes that the search index and archive are in sync
        if self.obj.target_url:
            return Origin(self.obj.target_url)
        return None


class OriginConnection(BaseConnection):
    """
    Connection resolver for the origins
    """

    _node_class = BaseOriginNode

    def _get_connection_data(self) -> ConnectionData:
        # Use the search backend if a urlPattern is given
        if self.kwargs.get("urlPattern"):
            origins = self.search.get_origins(
                query=self.kwargs.get("urlPattern"),
                after=self._get_after_arg(),
                first=self._get_first_arg(),
            )
            results = [Origin(ori["url"]) for ori in origins.results]
            paged_result = PagedResult(
                results=results, next_page_token=origins.next_page_token
            )
        else:
            # Use the archive backend by default
            paged_result = self.archive.get_origins(
                after=self._get_after_arg(), first=self._get_first_arg()
            )
        return ConnectionData(paged_result=paged_result)



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
from typing import Dict, Tuple

from ariadne import gql


def get_query_response(
    client, query_str: str, response_code: int = 200, **kwargs
) -> Tuple[Dict, Dict]:
    query = gql(query_str)
    response = client.post("/", json={"query": query, "variables": kwargs})
    assert response.status_code == response_code, response.content
    result = json.loads(response.content)
    return result.get("data"), result.get("errors")


def assert_missing_object(client, query_str: str, obj_type: str, **kwargs) -> None:
    data, errors = get_query_response(client, query_str, **kwargs)
    assert data[obj_type] is None
    assert len(errors) == 1
    assert errors[0]["message"] == "Object error: Requested object is not available"
    assert errors[0]["path"] == [obj_type]


def get_error_response(
    client, query_str: str, response_code: int = 200, **kwargs
) -> Dict:
    data, errors = get_query_response(
        client, query_str, response_code=response_code, **kwargs
    )
    assert len(errors) > 0
    return errors



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Optional, Union

from swh.model.model import CoreSWHID, ObjectType
from swh.model.model import Release as ModelRelease
from swh.model.model import Sha1Git

from .base_node import BaseSWHNode


class BaseReleaseNode(BaseSWHNode):
    """
    Base resolver for all the release nodes
    """

    def _get_release_by_id(self, release_id: Sha1Git) -> Optional[ModelRelease]:
        releases = self.archive.get_releases([release_id])
        return releases[0] if releases else None

    def target_hash(self) -> Optional[Sha1Git]:
        assert self._node is not None
        return self._node.target

    def target_type(self) -> ObjectType:
        assert self._node is not None
        return self._node.target_type

    def is_type_of(self) -> str:
        # is_type_of is required only when resolving a UNION type
        # This is for ariadne to return the right type
        return "Release"


class ReleaseNode(BaseReleaseNode):
    """
    Node resolver for a release requested directly with its SWHID
    """

    def _get_node_data(self) -> Optional[ModelRelease]:
        release_swhid = self.kwargs.get("swhid")
        assert isinstance(release_swhid, CoreSWHID)
        return self._get_release_by_id(release_swhid.object_id)


class TargetReleaseNode(BaseReleaseNode):
    """
    Node resolver for a release requested as a target
    """

    from .target import BranchTargetNode, TargetNode

    _can_be_null = True
    obj: Union[TargetNode, BranchTargetNode]

    def _get_node_data(self) -> Optional[ModelRelease]:
        # self.obj.target_hash is the requested release id
        return self._get_release_by_id(self.obj.target_hash)



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

import pytest

from swh.graphql.errors import InvalidInputError
from swh.graphql.resolvers import scalars


def test_serialize_id():
    assert scalars.serialize_id("test") == "74657374"
    assert scalars.serialize_id(b"test") == "74657374"


def test_validate_swhid_invalid():
    with pytest.raises(InvalidInputError):
        scalars.validate_swhid("invalid")


def test_validate_swhid():
    swhid = scalars.validate_swhid(f"swh:1:rev:{'1' * 40}")
    assert str(swhid) == "swh:1:rev:1111111111111111111111111111111111111111"


def test_serialize_datetime_from_datetime():
    dt = datetime.datetime(2010, 1, 15, 2, 12, 10, 2, datetime.timezone.utc)
    assert scalars.serialize_datetime(dt) == "2010-01-15T02:12:10.000002+00:00"


def test_serialize_datetime_invalid_input():
    assert scalars.serialize_datetime(None) is None



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import base64

import pytest

from swh.model.model import ObjectType

from . import utils
from ..data import (
    get_contents,
    get_directories,
    get_releases,
    get_releases_with_empty_target,
    get_releases_with_target,
    get_revisions,
)


@pytest.mark.parametrize("release", get_releases())
def test_get_release(client, release):
    query_str = """
    query getRelease($swhid: SWHID!) {
      release(swhid: $swhid) {
        swhid
        name {
          text
          base64
        }
        message {
          text
        }
        author {
          email {
            text
          }
          name {
            text
          }
          fullname {
            text
          }
        }
        date {
          date
          offset {
            text
            base64
          }
        }
        target {
          type
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=str(release.swhid()))

    assert data["release"] == {
        "swhid": str(release.swhid()),
        "name": {
            "text": release.name.decode(),
            "base64": base64.b64encode(release.name).decode("ascii"),
        },
        "message": {"text": release.message.decode()},
        "author": [
            {
                "email": {"text": release.author.email.decode()},
                "name": {"text": release.author.name.decode()},
                "fullname": {"text": release.author.fullname.decode()},
            }
        ]
        if release.author
        else [],
        "date": {
            "date": release.date.to_datetime().isoformat(),
            "offset": {
                "text": release.date.offset_bytes.decode(),
                "base64": base64.b64encode(release.date.offset_bytes).decode("ascii"),
            },
        }
        if release.date
        else None,
        "target": {"type": release.target_type.value},
    }


def test_get_release_with_invalid_swhid(client):
    query_str = """
    query getRelease($swhid: SWHID!) {
      release(swhid: $swhid) {
        swhid
      }
    }
    """
    errors = utils.get_error_response(client, query_str, swhid="swh:1:rel:invalid")
    # API will throw an error in case of an invalid SWHID
    assert len(errors) == 1
    assert "Input error: Invalid SWHID" in errors[0]["message"]


@pytest.mark.parametrize("release_with_target", get_releases_with_target())
def test_get_release_targets(client, release_with_target):
    query_str = """
    query getRelease($swhid: SWHID!) {
      release(swhid: $swhid) {
        target {
          type
          swhid
          node {
            ...on Revision {
              swhid
            }
            ...on Release {
              swhid
            }
            ...on Directory {
              swhid
            }
            ...on Content {
              swhid
            }
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(
        client, query_str, swhid=str(release_with_target.swhid())
    )

    if release_with_target.target_type == ObjectType.REVISION:
        target_swhid = get_revisions()[0].swhid()
    elif release_with_target.target_type == ObjectType.RELEASE:
        target_swhid = get_releases()[0].swhid()
    elif release_with_target.target_type == ObjectType.DIRECTORY:
        target_swhid = get_directories()[0].swhid()
    elif release_with_target.target_type == ObjectType.CONTENT:
        target_swhid = get_contents()[0].swhid()
    assert data["release"] == {
        "target": {
            "type": release_with_target.target_type.value,
            "swhid": str(target_swhid),
            "node": {"swhid": str(target_swhid)},
        }
    }


def test_get_release_target_unknown(client):
    # Clinet can request all the possible options if the target type
    # is unknown. The data under the right type will be returned

    # The target is of type Revision in this case
    # ie: both swhid and message will be available in the response
    swhid = get_releases_with_target()[0].swhid()
    query_str = """
    query getRelease($swhid: SWHID!) {
      release(swhid: $swhid) {
        target {
          type
          node {
            ...on Revision {
              swhid
              message {
                text
              }
            }
            ...on Release {
              swhid
            }
            ...on Directory {
              swhid
            }
            ...on Content {
              swhid
            }
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=str(swhid))
    assert data["release"] == {
        "target": {
            "type": "revision",
            "node": {
                "message": {"text": "hello"},
                "swhid": str(get_revisions()[0].swhid()),
            },
        },
    }


def test_get_release_with_unknown_swhid(client):
    unknown_sha1 = "1" * 40
    query_str = """
    query getRelease($swhid: SWHID!) {
      release(swhid: $swhid) {
        swhid
      }
    }
    """
    utils.assert_missing_object(
        client,
        query_str,
        obj_type="release",
        swhid=f"swh:1:rel:{unknown_sha1}",
    )


def test_get_release_with_empty_target(client):
    swhid = get_releases_with_empty_target()[0].swhid()
    query_str = """
    query getRelease($swhid: SWHID!) {
      release(swhid: $swhid) {
        target {
          type
        }
      }
    }"""
    data, err = utils.get_query_response(client, query_str, swhid=str(swhid))
    assert data == {"release": {"target": None}}
    assert err is None



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from .errors import (
    DataError,
    InvalidInputError,
    NullableObjectError,
    ObjectNotFoundError,
    PaginationError,
)
from .handlers import format_error, on_auth_error

__all__ = [
    "ObjectNotFoundError",
    "PaginationError",
    "InvalidInputError",
    "NullableObjectError",
    "DataError",
    "format_error",
    "on_auth_error",
]



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from graphql import GraphQLError, GraphQLSyntaxError
import pytest
import sentry_sdk

from swh.graphql import errors


def test_errors():
    err = errors.ObjectNotFoundError("test error")
    assert str(err) == "Object error: test error"

    err = errors.PaginationError("test error")
    assert str(err) == "Pagination error: test error"

    err = errors.InvalidInputError("test error")
    assert str(err) == "Input error: test error"


def test_format_error_with_debug():
    err = GraphQLError("test error")
    response = errors.format_error(err, debug=True)
    assert "extensions" in response


def test_format_error_without_debug():
    err = GraphQLError("test error")
    response = errors.format_error(err)
    assert "extensions" not in response


def test_format_error_sent_to_sentry(mocker):
    mocked_senty_call = mocker.patch.object(sentry_sdk, "capture_exception")
    err = GraphQLError("test error")
    err.original_error = NameError("test error")  # not an expected error
    errors.format_error(err)
    mocked_senty_call.assert_called_once_with(err)


@pytest.mark.parametrize(
    "error",
    [errors.ObjectNotFoundError, errors.PaginationError, errors.InvalidInputError],
)
def test_format_error_skip_sentry(mocker, error):
    mocked_senty_call = mocker.patch.object(sentry_sdk, "capture_exception")
    err = GraphQLError("test error")
    err.original_error = error("test error")
    errors.format_error(err)
    mocked_senty_call.assert_not_called()


def test_format_error_query_syntax_error_skip_sentry(mocker):
    mocked_senty_call = mocker.patch.object(sentry_sdk, "capture_exception")
    error = GraphQLSyntaxError(source=None, position=[1], description="test")
    errors.format_error(error)
    mocked_senty_call.assert_not_called()



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

from .base_connection import BaseConnection, ConnectionData
from .base_node import BaseNode


class OriginSearchResultNode(BaseNode):

    obj: OriginSearchConnection

    def _get_node_from_data(self, node_data: dict):
        # overriding to enrich the node_data returned
        # by OriginSearchConnection._get_connection_data
        updated_node_data = {
            # Field exposed in the schema
            "type": "origin",
            # Field exposed in the schema
            "url": node_data.get("url"),
            # Field NOT exposed in the schema, used to get the target node
            "target_url": node_data.get("url"),
        }
        return super()._get_node_from_data(updated_node_data)


class OriginSearchConnection(BaseConnection):

    _node_class = OriginSearchResultNode

    def _get_connection_data(self) -> ConnectionData:
        origins = self.search.get_origins(
            query=self.kwargs.get("query"),
            after=self._get_after_arg(),
            first=self._get_first_arg(),
        )
        return ConnectionData(paged_result=origins)



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import TYPE_CHECKING, Optional, Union

from swh.graphql.errors import NullableObjectError
from swh.model.model import Directory

from .base_node import BaseSWHNode
from .revision import BaseRevisionNode


class BaseDirectoryNode(BaseSWHNode):
    """
    Base resolver for all the directory nodes
    """

    def is_type_of(self):
        return "Directory"


class DirectoryNode(BaseDirectoryNode):
    """
    Node resolver for a directory requested directly with its SWHID
    """

    def _get_node_data(self):
        swhid = self.kwargs.get("swhid")
        return self.archive.get_directory(directory_id=swhid.object_id, verify=True)


class RevisionDirectoryNode(BaseDirectoryNode):
    """
    Node resolver for a directory requested from a revision
    """

    # Revision directory is not using the target indirection,
    # hence this implementation (not using TargetDirectoryNode)

    _can_be_null = True
    obj: BaseRevisionNode

    def _get_node_data(self) -> Directory:
        # self.obj.directory_hash is the requested directory Id
        directory_id = self.obj.directory_hash()
        if directory_id is None:
            raise NullableObjectError()
        directory = self.archive.get_directory(directory_id=directory_id, verify=False)
        assert directory is not None
        return directory


class TargetDirectoryNode(BaseDirectoryNode):
    """
    Node resolver for a directory requested as a target
    """

    if TYPE_CHECKING:  # pragma: no cover
        from .target import BranchTargetNode, TargetNode

        obj: Union[
            BranchTargetNode,
            TargetNode,
        ]
    _can_be_null = True

    def _get_node_data(self) -> Optional[Directory]:
        # existing directory in the archive, hence verify is False
        return self.archive.get_directory(
            directory_id=self.obj.target_hash, verify=False
        )



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# import pkg_resources
import os
from pathlib import Path

from ariadne import gql, load_schema_from_path, make_executable_schema
from ariadne.validation import cost_validator

from .resolvers import resolvers, scalars

type_defs = gql(
    # pkg_resources.resource_string("swh.graphql", "schema/schema.graphql").decode()
    load_schema_from_path(
        os.path.join(Path(__file__).parent.resolve(), "schema", "schema.graphql")
    )
)

schema = make_executable_schema(
    type_defs,
    resolvers.query,
    resolvers.origin,
    resolvers.visit,
    resolvers.visit_status,
    resolvers.snapshot,
    resolvers.snapshot_branch,
    resolvers.revision,
    resolvers.release,
    resolvers.directory,
    resolvers.directory_entry,
    resolvers.branch_target,
    resolvers.branch_target_node,
    resolvers.release_target,
    resolvers.release_target_node,
    resolvers.directory_entry_target,
    resolvers.directory_entry_target_node,
    resolvers.resolve_swhid_result,
    resolvers.origin_search_result,
    resolvers.binary_string,
    resolvers.date,
    scalars.id_scalar,
    scalars.datetime_scalar,
    scalars.swhid_scalar,
)


def validation_rules(context, document, data):
    from .server import graphql_cfg

    if context["request"].user.is_authenticated:
        max_query_cost = graphql_cfg["max_query_cost"]["user"]
    else:
        max_query_cost = graphql_cfg["max_query_cost"]["anonymous"]

    if max_query_cost:
        return [
            cost_validator(maximum_cost=max_query_cost, variables=data.get("variables"))
        ]
    # no limit is applied when max_query_cost is set to 0 or None
    return None



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Optional

from swh.graphql.utils import utils
from swh.model.model import Sha1Git

from .base_connection import BaseConnection, ConnectionData
from .base_node import BaseNode
from .visit import BaseVisitNode


class BaseVisitStatusNode(BaseNode):
    """
    Base resolver for all the visit-status nodes
    """

    def snapshot_id(self) -> Optional[Sha1Git]:
        assert self._node is not None
        if self._node.snapshot is None:
            return None
        return self._node.snapshot


class LatestVisitStatusNode(BaseVisitStatusNode):
    """
    Node resolver for a visit-status requested from a visit
    """

    _can_be_null = True
    obj: BaseVisitNode

    def _get_node_data(self):
        # self.obj.origin is the origin URL
        return self.archive.get_latest_visit_status(
            origin_url=self.obj.origin,
            visit_id=self.obj.visitId,
            allowed_statuses=self.kwargs.get("allowedStatuses"),
            require_snapshot=self.kwargs.get("requireSnapshot"),
        )


class VisitStatusConnection(BaseConnection):
    """
    Connection resolver for the visit-status objects in a visit
    """

    obj: BaseVisitNode
    _node_class = BaseVisitStatusNode

    def _get_connection_data(self) -> ConnectionData:
        # self.obj.origin is the origin URL
        return ConnectionData(
            paged_result=self.archive.get_visit_statuses(
                self.obj.origin,
                self.obj.visitId,
                after=self._get_after_arg(),
                first=self._get_first_arg(),
                order=utils.get_storage_list_order(self.kwargs.get("sort")),
            )
        )

    def _get_index_cursor(self, index: int, node: BaseVisitStatusNode):
        # Visit status is using a different cursor, hence the override
        return utils.get_encoded_cursor(utils.get_formatted_date(node.date))



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from . import utils
from ..data import get_directories, get_origins


# Using Origin object to run functional tests for pagination
def get_origin_nodes(client, first, after=""):
    query_str = """
    query getOrigins($first: Int!, $after: String) {
      origins(first: $first, after: $after) {
        nodes {
          id
        }
        pageInfo {
          hasNextPage
          endCursor
        }
      }
    }
    """
    return utils.get_query_response(client, query_str, first=first, after=after)


def test_pagination(client):
    # requesting the max number of nodes available
    # endCursor must be None
    data, _ = get_origin_nodes(client, first=len(get_origins()))
    assert len(data["origins"]["nodes"]) == len(get_origins())
    assert data["origins"]["pageInfo"] == {"hasNextPage": False, "endCursor": None}


def test_first_arg(client):
    data, _ = get_origin_nodes(client, first=1)
    assert len(data["origins"]["nodes"]) == 1
    assert data["origins"]["pageInfo"]["hasNextPage"] is True


def test_invalid_first_arg(client):
    data, errors = get_origin_nodes(client, first=-1)
    assert data["origins"] is None
    assert (len(errors)) == 2  # one error for origins and anotehr one for pageInfo
    assert (
        errors[0]["message"]
        == "Pagination error: Value for argument 'first' is invalid; it must be between 0 and 1000"  # noqa: B950
    )


def test_too_big_first_arg(client, high_query_cost):
    data, errors = get_origin_nodes(client, 1001)  # max page size is 1000
    assert data["origins"] is None
    assert (len(errors)) == 2
    assert (
        errors[0]["message"]
        == "Pagination error: Value for argument 'first' is invalid; it must be between 0 and 1000"  # noqa: B950
    )


def test_after_arg(client):
    first_data, _ = get_origin_nodes(client, first=1)
    end_cursor = first_data["origins"]["pageInfo"]["endCursor"]
    # get again with endcursor as the after argument
    data, _ = get_origin_nodes(client, first=1, after=end_cursor)
    assert len(data["origins"]["nodes"]) == 1
    assert data["origins"]["pageInfo"] == {"hasNextPage": False, "endCursor": None}


def test_invalid_after_arg(client):
    data, errors = get_origin_nodes(client, first=1, after="invalid")
    assert data["origins"] is None
    assert (len(errors)) == 2
    assert (
        errors[0]["message"] == "Pagination error: Invalid value for argument 'after'"
    )


def test_edge_cursor(client):
    # Use DirectoryEntry connection to test connection edges
    # The same code is used in all the other connections, hence a single test
    directory = get_directories()[1]
    query_str = """
    query getDirectory($swhid: SWHID!, $entries_first: Int!, $entries_after: String) {
      directory(swhid: $swhid) {
        entries(first: $entries_first, after: $entries_after) {
          edges {
            cursor
            node {
              name {
                text
              }
            }
          }
          pageInfo {
            endCursor
            hasNextPage
          }
          nodes {
             name {
               text
             }
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(
        client, query_str, swhid=str(directory.swhid()), entries_first=1
    )
    entries = data["directory"]["entries"]
    # Make sure the first node object in edges and nodes is the same
    assert entries["edges"][0]["node"] == entries["nodes"][0]
    assert entries["edges"][0]["node"] == {
        "name": {"text": directory.entries[1].name.decode()}
    }
    assert entries["pageInfo"]["hasNextPage"] is True
    # FIXME, Following behaviour is not in compliance with the relay spec.
    # last-item-cursor and endcursor should be the same as per relay.
    # This test will fail once the pagination becomes fully relay complaint.
    assert entries["pageInfo"]["endCursor"] != entries["edges"][-1]["cursor"]
    # Make another query with the after argument, after argument is the first item
    # cursor here, result will be the same as the last one
    new_data, _ = utils.get_query_response(
        client,
        query_str,
        swhid=str(directory.swhid()),
        entries_first=1,
        entries_after=entries["edges"][0]["cursor"],
    )
    assert new_data == data
    # Make another query with the end cursor from the first query
    final_data, _ = utils.get_query_response(
        client,
        query_str,
        swhid=str(directory.swhid()),
        entries_first=2,
        entries_after=entries["pageInfo"]["endCursor"],
    )
    final_result_entries = final_data["directory"]["entries"]
    # endcursor from the first query will be the first item cursor here
    # FIXME, this behaviour is not in compliance with the relay spec.
    # With relay spec, items after the given cursor ($after) will be returned
    assert (
        final_result_entries["edges"][0]["cursor"] == entries["pageInfo"]["endCursor"]
    )
    assert final_result_entries["nodes"] == [
        {"name": {"text": directory.entries[0].name.decode()}},
        {"name": {"text": directory.entries[2].name.decode()}},
    ]



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.graphql import server
from swh.model.swhids import CoreSWHID, ObjectType

from . import utils
from ..data import (
    get_directories,
    get_directories_with_nested_path,
    get_directories_with_special_name_entries,
)


def get_target_type(target_type):
    mapping = {"file": "content", "dir": "directory", "rev": "revision"}
    return mapping.get(target_type)


def test_get_directory_entry_missing_path(client):
    directory = get_directories()[0]
    path = "missing"
    query_str = """
    query getDirEntry($swhid: SWHID!, $path: String!) {
      directoryEntry(directorySWHID: $swhid, path: $path) {
        name {
          text
        }
        target {
          type
          swhid
          node {
            ...on Content {
              swhid
            }
          }
        }
      }
    }
    """
    utils.assert_missing_object(
        client,
        query_str,
        "directoryEntry",
        swhid=str(directory.swhid()),
        path=path,
    )


@pytest.mark.parametrize(
    "directory", get_directories() + get_directories_with_nested_path()
)
def test_get_directory_entry(client, directory):
    storage = server.get_storage()
    query_str = """
    query getDirEntry($swhid: SWHID!, $path: String!) {
      directoryEntry(directorySWHID: $swhid, path: $path) {
        name {
          text
        }
        target {
          type
          swhid
          node {
            ...on Content {
              swhid
            }
            ...on Directory {
              swhid
            }
            ...on Revision {
              swhid
            }
          }
        }
      }
    }
    """
    for entry in storage.directory_ls(directory.id, recursive=True):
        data, _ = utils.get_query_response(
            client,
            query_str,
            swhid=str(directory.swhid()),
            path=entry["name"].decode(),
        )
        swhid = None
        node_exists = True
        if entry["type"] == "file" and entry["target"] is not None:
            swhid = CoreSWHID(object_type=ObjectType.CONTENT, object_id=entry["target"])
        elif entry["type"] == "dir" and entry["target"] is not None:
            swhid = CoreSWHID(
                object_type=ObjectType.DIRECTORY, object_id=entry["target"]
            )
        elif entry["type"] == "rev" and entry["target"] is not None:
            swhid = CoreSWHID(
                object_type=ObjectType.REVISION, object_id=entry["target"]
            )
        if entry["target"] == b"\x11" * 20:
            # This is a non existing object, just the target reference exists
            node_exists = False
        assert data["directoryEntry"] == {
            "name": {"text": entry["name"].decode()},
            "target": {
                "type": get_target_type(entry["type"]),
                "swhid": str(swhid) if swhid else None,
                "node": {"swhid": str(swhid)} if (swhid and node_exists) else None,
            },
        }


@pytest.mark.parametrize("directory", get_directories())
def test_get_directory_entry_connection(client, directory):
    query_str = """
    query getDirectory($swhid: SWHID!) {
      directory(swhid: $swhid) {
        swhid
        entries(first: 10) {
          totalCount
          nodes {
            name {
              text
            }
            target {
              type
            }
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=str(directory.swhid()))
    directory_entries = data["directory"]["entries"]["nodes"]
    assert len(directory_entries) == len(directory.entries)
    assert data["directory"]["entries"]["totalCount"] == len(directory.entries)
    output = [
        {
            "name": {"text": de.name.decode()},
            "target": {"type": get_target_type(de.type)},
        }
        for de in directory.entries
    ]
    for each_entry in output:
        assert each_entry in directory_entries


@pytest.mark.parametrize("directory", get_directories())
def test_directory_entry_connection_filter_by_name(client, directory):
    storage = server.get_storage()
    for dir_entry in storage.directory_ls(directory.id):
        name_include = dir_entry["name"][:-1].decode()
        query_str = """
        query getDirectory($swhid: SWHID!, $nameInclude: String) {
          directory(swhid: $swhid) {
            swhid
            entries(nameInclude: $nameInclude) {
              nodes {
                name {
                  text
                }
                target {
                  type
                }
              }
            }
          }
        }
        """
        data, _ = utils.get_query_response(
            client,
            query_str,
            swhid=str(directory.swhid()),
            nameInclude=name_include,
        )
        for entry in data["directory"]["entries"]["nodes"]:
            assert name_include in entry["name"]["text"]
            assert entry["target"]["type"] == get_target_type(dir_entry["type"])


def test_directory_entry_connection_filter_by_name_special_chars(client):
    directory = get_directories_with_special_name_entries()[0]
    query_str = """
    query getDirectory($swhid: SWHID!, $nameInclude: String) {
      directory(swhid: $swhid) {
        entries(nameInclude: $nameInclude) {
          nodes {
            name {
              text
            }
            target {
              type
            }
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(
        client,
        query_str,
        swhid=str(directory.swhid()),
        nameInclude="ssSSé",
    )
    assert data["directory"]["entries"]["nodes"][0] == {
        "name": {"text": "ßßétEÉt"},
        "target": {
            "type": "content",
        },
    }



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import base64
from datetime import datetime
from typing import TYPE_CHECKING, List, Optional

from swh.graphql.errors import InvalidInputError

if TYPE_CHECKING:
    from swh.graphql.resolvers.base_connection import ConnectionData

from swh.storage.interface import ListOrder, PagedResult

ENCODING = "utf-8"


def get_b64_string(source) -> str:
    if isinstance(source, str):
        source = source.encode(ENCODING)
    return base64.b64encode(source).decode("ascii")


def get_encoded_cursor(cursor: Optional[str]) -> Optional[str]:
    if cursor is None:
        return None
    return get_b64_string(cursor)


def get_decoded_cursor(cursor: Optional[str]) -> Optional[str]:
    if cursor is None:
        return None
    return base64.b64decode(cursor, validate=True).decode()


def get_formatted_date(date: datetime) -> str:
    # FIXME, handle error + return other formats
    return date.isoformat()


def get_storage_list_order(order: str) -> ListOrder:
    mapping = {"ASC": ListOrder.ASC, "DESC": ListOrder.DESC}
    if order not in mapping:
        raise InvalidInputError("Invalid sort order")
    return mapping[order]


def get_local_paginated_data(source: List, first: int, after=0) -> "ConnectionData":
    """
    Pagination at the GraphQL level
    This is a temporary fix and inefficient. Should eventually be moved to the
    backend (storage) level
    """
    from swh.graphql.resolvers.base_connection import ConnectionData

    # FIXME, handle data errors here
    after = 0 if after is None else int(after)
    end_cursor = after + first
    results = source[after:end_cursor]
    next_page_token = None
    if len(source) > end_cursor:
        next_page_token = str(end_cursor)
    return ConnectionData(
        paged_result=PagedResult(results=results, next_page_token=next_page_token),
        total_count=len(source),
    )



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.graphql import server


def test_load_and_check_config_no_config():
    with pytest.raises(EnvironmentError):
        server.load_and_check_config(config_path=None)


def test_load_and_check_config_missing_config_file():
    with pytest.raises(FileNotFoundError):
        server.load_and_check_config(config_path="invalid")


def test_load_and_check_config_missing_storage_config(mocker):
    mocker.patch("swh.core.config.read", return_value={"test": "test"})
    with pytest.raises(KeyError):
        server.load_and_check_config(config_path="/tmp")


def test_load_and_check_config(mocker):
    mocker.patch("swh.core.config.read", return_value={"storage": {"test": "test"}})
    cfg = server.load_and_check_config(config_path="/tmp")
    assert cfg == {"storage": {"test": "test"}}



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import base64

import pytest

from swh.model.swhids import CoreSWHID

from . import utils
from ..data import (
    get_revisions,
    get_revisions_with_none_date,
    get_revisions_with_parents,
)


@pytest.mark.parametrize("revision", get_revisions())
def test_get_revision(client, revision):
    query_str = """
    query getRevision($swhid: SWHID!) {
      revision(swhid: $swhid) {
        swhid
        message {
          text
        }
        author {
          fullname {
            text
          }
          name {
            text
          }
          email {
            text
          }
        }
        committer {
          fullname {
            text
          }
          name {
            text
          }
          email {
            text
          }
        }
        date {
          date
          offset {
            text
            base64
          }
        }
        committerDate {
          date
          offset {
            text
            base64
          }
        }
        type
        directory {
          swhid
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=str(revision.swhid()))

    assert data["revision"] == {
        "swhid": str(revision.swhid()),
        "message": {"text": revision.message.decode()},
        "author": [
            {
                "fullname": {"text": revision.author.fullname.decode()},
                "name": {"text": revision.author.name.decode()},
                "email": {"text": revision.author.email.decode()},
            }
        ],
        "committer": [
            {
                "fullname": {"text": revision.committer.fullname.decode()},
                "name": {"text": revision.committer.name.decode()},
                "email": {"text": revision.committer.email.decode()},
            }
        ],
        "date": {
            "date": revision.date.to_datetime().isoformat(),
            "offset": {
                "text": revision.date.offset_bytes.decode(),
                "base64": base64.b64encode(revision.date.offset_bytes).decode("ascii"),
            },
        }
        if revision.date
        else None,
        "committerDate": {
            "date": revision.committer_date.to_datetime().isoformat(),
            "offset": {
                "text": revision.committer_date.offset_bytes.decode(),
                "base64": base64.b64encode(revision.committer_date.offset_bytes).decode(
                    "ascii"
                ),
            },
        }
        if revision.committer_date
        else None,
        "type": revision.type.value,
        "directory": {
            "swhid": str(CoreSWHID(object_id=revision.directory, object_type="dir")),
        },
    }


def test_get_revision_with_invalid_swhid(client):
    query_str = """
    query getRevision($swhid: SWHID!) {
      revision(swhid: $swhid) {
        swhid
      }
    }
    """
    errors = utils.get_error_response(client, query_str, swhid="swh:1:cnt:invalid")
    # API will throw an error in case of an invalid SWHID
    assert len(errors) == 1
    assert "Input error: Invalid SWHID" in errors[0]["message"]


def test_get_revision_as_target(client):
    # SWHID of a snapshot with revision as target
    snapshot_swhid = "swh:1:snp:9e78d7105c5e0f886487511e2a92377b4ee4c32a"
    query_str = """
    query getSnapshot($swhid: SWHID!) {
      snapshot(swhid: $swhid) {
        branches(first: 1, types: [revision]) {
          nodes {
            target {
              node {
                ...on Revision {
                  swhid
                }
              }
            }
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=snapshot_swhid)
    revision_obj = data["snapshot"]["branches"]["nodes"][0]["target"]["node"]
    assert revision_obj == {
        "swhid": "swh:1:rev:66c7c1cd9673275037140f2abff7b7b11fc9439c"
    }


def test_get_revision_log(client):
    revision_swhid = get_revisions_with_parents()[0].swhid()
    query_str = """
    query getRevision($swhid: SWHID!) {
      revision(swhid: $swhid) {
        swhid
        revisionLog(first: 2) {
          totalCount
          nodes {
            swhid
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=str(revision_swhid))
    assert data["revision"]["revisionLog"] == {
        "totalCount": 3,
        "nodes": [
            {"swhid": str(revision_swhid)},
            {"swhid": str(get_revisions()[0].swhid())},
        ],
    }


def test_get_revision_parents(client):
    revision_swhid = get_revisions_with_parents()[0].swhid()
    query_str = """
    query getRevision($swhid: SWHID!) {
      revision(swhid: $swhid) {
        swhid
        parents {
          totalCount
          nodes {
            swhid
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=str(revision_swhid))

    assert data["revision"]["parents"] == {
        "totalCount": 2,
        "nodes": [
            {"swhid": str(get_revisions()[0].swhid())},
            {"swhid": str(get_revisions()[1].swhid())},
        ],
    }


def test_get_revision_with_unknown_swhid(client):
    unknown_sha1 = "1" * 40
    query_str = """
    query getRevision($swhid: SWHID!) {
      revision(swhid: $swhid) {
        swhid
      }
    }
    """
    utils.assert_missing_object(
        client,
        query_str,
        obj_type="revision",
        swhid=f"swh:1:rev:{unknown_sha1}",
    )


def test_get_revisions_with_none_date(client):
    revision_swhid = get_revisions_with_none_date()[0].swhid()
    query_str = """
    query getRevision($swhid: SWHID!) {
      revision(swhid: $swhid) {
        swhid
        date {
          date
          offset {
            text
            base64
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=str(revision_swhid))
    assert data == {"revision": {"swhid": str(revision_swhid), "date": None}}



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from . import utils
from ..data import get_directories


@pytest.mark.parametrize("directory", get_directories())
def test_get_directory(client, directory):
    query_str = """
    query getDirectory($swhid: SWHID!) {
      directory(swhid: $swhid) {
        swhid
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=str(directory.swhid()))
    assert data["directory"] == {"swhid": str(directory.swhid())}


def test_get_directory_with_invalid_swhid(client):
    query_str = """
    query getDirectory($swhid: SWHID!) {
      directory(swhid: $swhid) {
        swhid
      }
    }
    """
    errors = utils.get_error_response(client, query_str, swhid="swh:1:dir:invalid")
    # API will throw an error in case of an invalid SWHID
    assert len(errors) == 1
    assert "Input error: Invalid SWHID" in errors[0]["message"]


def test_get_revision_directory(client):
    query_str = """
    query getRevision($swhid: SWHID!) {
      revision(swhid: $swhid) {
        swhid
        directory {
          swhid
        }
      }
    }
    """
    data, _ = utils.get_query_response(
        client,
        query_str,
        swhid="swh:1:rev:66c7c1cd9673275037140f2abff7b7b11fc9439c",
    )
    assert data["revision"]["directory"] == {
        "swhid": "swh:1:dir:0101010101010101010101010101010101010101",
    }


def test_get_target_directory(client):
    # TargetDirectoryNode is returned from snapshotbranch, release
    # and directory entry nodes. Release node is used for testing here
    query_str = """
    query getRelease($swhid: SWHID!) {
      release(swhid: $swhid) {
        swhid
        target {
          node {
            ...on Directory {
              swhid
            }
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(
        client,
        query_str,
        swhid="swh:1:rel:ee4d20e80af850cc0f417d25dc5073792c5010d2",
    )
    assert data["release"]["target"]["node"] == {
        "swhid": "swh:1:dir:0505050505050505050505050505050505050505"
    }


def test_get_directory_with_unknown_swhid(client):
    unknown_sha1 = "1" * 40
    query_str = """
    query getDirectory($swhid: SWHID!) {
      directory(swhid: $swhid) {
        swhid
      }
    }
    """
    utils.assert_missing_object(
        client,
        query_str,
        obj_type="directory",
        swhid=f"swh:1:dir:{unknown_sha1}",
    )



# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information
from unittest.mock import ANY, call

from swh.core import statsd

from . import utils
from ..data import get_directories, get_origins


def test_node_usage_logger(client, mocker):
    statsd_report = mocker.patch.object(statsd.Statsd, "_report")
    query_str = """
    query getOrigin($url: String!) {
      origin(url: $url) {
        url
      }
    }
    """
    response, _ = utils.get_query_response(client, query_str, url=get_origins()[0].url)
    statsd_calls = statsd_report.mock_calls
    assert statsd_calls == [
        # total count
        call("query_total", "c", 1, None, 1),
        # node duration
        call("node_query_seconds", "ms", ANY, {"node": "origin"}, 1),
        # total query  duration
        call("query_seconds", "ms", ANY, {}, 1),
    ]


def test_connection_usage_logger(client, mocker):
    statsd_report = mocker.patch.object(statsd.Statsd, "_report")
    query_str = """
    query getOrigins($first: Int!) {
      origins(first: $first) {
        nodes {
          url
        }
      }
    }
    """
    response, _ = utils.get_query_response(client, query_str, first=10)
    statsd_calls = statsd_report.mock_calls
    assert statsd_calls == [
        # total count
        call("query_total", "c", 1, None, 1),
        # connection duration
        call("connection_query_seconds", "ms", ANY, {"connection": "origins"}, 1),
        # total query  duration
        call("query_seconds", "ms", ANY, {}, 1),
    ]


def test_list_usage_logger(client, mocker):
    statsd_report = mocker.patch.object(statsd.Statsd, "_report")
    query_str = """
    query resolve($swhid: SWHID!) {
      resolveSWHID(swhid: $swhid) {
        ... on Directory {
          swhid
        }
      }
    }
    """
    response, _ = utils.get_query_response(
        client, query_str, swhid=str(get_directories()[0].swhid())
    )
    statsd_calls = statsd_report.mock_calls
    assert statsd_calls == [
        # total count
        call("query_total", "c", 1, None, 1),
        # list duration
        call("list_query_seconds", "ms", ANY, {"list": "resolve-swhid"}, 1),
        # total query  duration
        call("query_seconds", "ms", ANY, {}, 1),
    ]


def test_multiple_entrypoints_query_logger(client, mocker):
    statsd_report = mocker.patch.object(statsd.Statsd, "_report")
    query_str = """
    query big($url: String!, $first: Int!) {
      origins(first: $first) {
        nodes {
          url
        }
      }

      origin(url: $url) {
        url
      }
    }
    """
    response, _ = utils.get_query_response(
        client, query_str, url=get_origins()[0].url, first=2
    )
    statsd_calls = statsd_report.mock_calls
    assert statsd_calls == [
        # total count
        call("query_total", "c", 1, None, 1),
        ANY,
        ANY,
        # total query  duration
        call("query_seconds", "ms", ANY, {}, 1),
    ]
    # call order could change after moving to async
    assert call("node_query_seconds", "ms", ANY, {"node": "origin"}, 1) in statsd_calls
    assert (
        call("connection_query_seconds", "ms", ANY, {"connection": "origins"}, 1)
        in statsd_calls
    )



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

from swh.model.model import (
    Directory,
    DirectoryEntry,
    ObjectType,
    OriginVisitStatus,
    Release,
    Revision,
    RevisionType,
    Snapshot,
    SnapshotBranch,
    TargetType,
)
from swh.model.tests import swh_model_data

UTC = datetime.timezone.utc


def populate_search_data(search):
    search.origin_update({"url": origin.url} for origin in get_origins())


def get_origins():
    return swh_model_data.ORIGINS


def get_visits():
    return swh_model_data.ORIGIN_VISITS


def get_visit_status():
    return swh_model_data.ORIGIN_VISIT_STATUSES


def get_snapshots():
    return swh_model_data.SNAPSHOTS


def get_releases():
    return swh_model_data.RELEASES


def get_revisions():
    return swh_model_data.REVISIONS


def get_contents():
    return swh_model_data.CONTENTS


def get_directories():
    return swh_model_data.DIRECTORIES


def get_releases_with_target():
    """
    GraphQL will not return a target object unless the target id
    is present in the DB.
    Return release objects with real targets instead of dummy
    targets in swh.model.tests.swh_model_data
    """
    with_revision = Release(
        name=b"v0.0.1",
        target_type=ObjectType.REVISION,
        target=get_revisions()[0].id,
        message=b"foo",
        synthetic=False,
    )
    with_release = Release(
        name=b"v0.0.1",
        target_type=ObjectType.RELEASE,
        target=get_releases()[0].id,
        message=b"foo",
        synthetic=False,
    )
    with_directory = Release(
        name=b"v0.0.1",
        target_type=ObjectType.DIRECTORY,
        target=get_directories()[0].id,
        message=b"foo",
        synthetic=False,
    )
    with_content = Release(
        name=b"v0.0.1",
        target_type=ObjectType.CONTENT,
        target=get_contents()[0].sha1_git,
        message=b"foo",
        synthetic=False,
    )
    return [with_revision, with_release, with_directory, with_content]


def get_releases_with_empty_target():
    return [
        Release(
            name=b"v0.0.1",
            target_type=ObjectType.REVISION,
            target=b"",
            message=b"foo",
            synthetic=False,
        )
    ]


def get_revisions_with_parents():
    """
    Revisions with real revisions as parents
    """
    return [
        Revision(
            message=b"hello",
            date=swh_model_data.DATES[0],
            committer=swh_model_data.COMMITTERS[0],
            author=swh_model_data.COMMITTERS[0],
            committer_date=swh_model_data.DATES[0],
            type=RevisionType.GIT,
            directory=b"\x01" * 20,
            synthetic=False,
            parents=(get_revisions()[0].id, get_revisions()[1].id),
        )
    ]


def get_revisions_with_none_date():
    return [
        Revision(
            message=b"hello",
            date=None,
            committer=swh_model_data.COMMITTERS[0],
            author=swh_model_data.COMMITTERS[0],
            committer_date=swh_model_data.DATES[0],
            type=RevisionType.GIT,
            directory=b"\x01" * 20,
            synthetic=False,
            parents=(get_revisions()[0].id, get_revisions()[1].id),
        )
    ]


def get_directories_with_nested_path():
    return [
        Directory(
            entries=(
                DirectoryEntry(
                    name=b"sub-dir",
                    perms=0o644,
                    type="dir",
                    target=get_directories()[1].id,
                ),
            )
        )
    ]


def get_directories_with_special_name_entries():
    return [
        Directory(
            entries=(
                DirectoryEntry(
                    name="ßßétEÉt".encode(),
                    perms=0o644,
                    type="file",
                    target=get_contents()[0].sha1_git,
                ),
            )
        )
    ]


def get_visit_with_multiple_status():
    return [
        OriginVisitStatus(
            origin=get_origins()[0].url,
            date=datetime.datetime(2014, 5, 7, 4, 20, 39, 432222, tzinfo=UTC),
            visit=1,
            type="git",
            status="ongoing",
            snapshot=None,
            metadata=None,
        )
    ]


def get_snapshots_with_multiple_alias():
    return [
        Snapshot(
            # This branch alias chain breaks without a target
            branches={
                b"target/alias1": SnapshotBranch(
                    target_type=TargetType.ALIAS, target=b"target/alias2"
                ),
                b"target/alias2": SnapshotBranch(
                    target_type=TargetType.ALIAS, target=b"target/alias3"
                ),
            },
        ),
        Snapshot(
            # This branch alias chain resolves to a release after 2 levels
            branches={
                b"target/alias1": SnapshotBranch(
                    target_type=TargetType.ALIAS, target=b"target/alias2"
                ),
                b"target/alias2": SnapshotBranch(
                    target_type=TargetType.ALIAS, target=b"target/release"
                ),
                b"target/release": SnapshotBranch(
                    target_type=TargetType.RELEASE, target=get_releases()[0].id
                ),
            },
        ),
        Snapshot(
            # This branch alias chain is going 6 levels deep
            branches={
                b"target/alias1": SnapshotBranch(
                    target_type=TargetType.ALIAS, target=b"target/alias2"
                ),
                b"target/alias2": SnapshotBranch(
                    target_type=TargetType.ALIAS, target=b"target/alias3"
                ),
                b"target/alias3": SnapshotBranch(
                    target_type=TargetType.ALIAS, target=b"target/alias4"
                ),
                b"target/alias4": SnapshotBranch(
                    target_type=TargetType.ALIAS, target=b"target/alias5"
                ),
                b"target/alias5": SnapshotBranch(
                    target_type=TargetType.ALIAS, target=b"target/alias6"
                ),
                b"target/alias6": SnapshotBranch(
                    target_type=TargetType.ALIAS, target=b"target/release"
                ),
                b"target/release": SnapshotBranch(
                    target_type=TargetType.RELEASE, target=get_releases()[0].id
                ),
            },
        ),
        Snapshot(
            # This branch alias chain is recursive
            branches={
                b"target/alias1": SnapshotBranch(
                    target_type=TargetType.ALIAS, target=b"target/alias2"
                ),
                b"target/alias2": SnapshotBranch(
                    target_type=TargetType.ALIAS, target=b"target/alias1"
                ),
            },
        ),
    ]


GRAPHQL_EXTRA_TEST_OBJECTS = {
    "snapshot": get_snapshots_with_multiple_alias(),
    "release": get_releases_with_target() + get_releases_with_empty_target(),
    "revision": get_revisions_with_parents() + get_revisions_with_none_date(),
    "directory": get_directories_with_nested_path()
    + get_directories_with_special_name_entries(),
    "origin_visit_status": get_visit_with_multiple_status(),
}


def populate_dummy_data(storage):
    for object_type, objects in swh_model_data.TEST_OBJECTS.items():
        method = getattr(storage, object_type + "_add")
        method(objects)
    for object_type, objects in GRAPHQL_EXTRA_TEST_OBJECTS.items():
        method = getattr(storage, object_type + "_add")
        method(objects)



# Copyright (C) 2022-2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import TYPE_CHECKING, Dict, Optional, Union

from swh.model.model import CoreSWHID
from swh.model.swhids import ObjectType as SwhidObjectType

from .base_node import BaseNode
from .snapshot_branch import SnapshotBranchNode


class BaseTargetNode(BaseNode):

    # 'node' field in this object is resolved in the top level

    @property
    def swhid(self) -> Optional[CoreSWHID]:
        # field exposed in the schema
        # use the target type and hash to construct the SWHID
        mapping = {  # to map models.ObjectId to swhids.ObjectId
            "snapshot": SwhidObjectType.SNAPSHOT,
            "revision": SwhidObjectType.REVISION,
            "release": SwhidObjectType.RELEASE,
            "directory": SwhidObjectType.DIRECTORY,
            "content": SwhidObjectType.CONTENT,
        }
        if self.target_hash and self.type:
            return CoreSWHID(object_type=mapping[self.type], object_id=self.target_hash)
        return None


class TargetNode(BaseTargetNode):
    """
    Intermediate node between an object and its target
    Created for schema clarity and to handle missing target
    nodes in the archive
    """

    if TYPE_CHECKING:  # pragma: no cover
        from .directory_entry import BaseDirectoryEntryNode
        from .release import BaseReleaseNode
        from .revision import BaseRevisionNode

        obj: Union[BaseReleaseNode, BaseDirectoryEntryNode, BaseRevisionNode]

    _can_be_null = True

    def _get_node_data(self) -> Optional[Dict]:
        # No exta data to fetch; everything is available from self.obj
        if not self.obj.target_hash():
            # No information to load the target, consider it as a None target
            return None
        return {
            # field exposed in the schema for some nodes
            "type": self.obj.target_type().value,
            # field NOT exposed in the schema
            # to be used while retrieving the node object
            "target_hash": self.obj.target_hash(),
        }


class BranchTargetNode(BaseTargetNode):
    # Return the final target and the chain

    obj: SnapshotBranchNode

    def _get_node_data(self) -> Dict:
        branch, chain = self.archive.get_branch_target(
            snapshot_id=self.obj.snapshot_id,
            branch_obj=self.obj.branch_obj,
            max_length=5,
        )
        return {
            # field exposed in the schema, return None instead of an empty list
            "resolveChain": chain if chain else None,
            # field exposed in the schema
            "type": branch.target_type.value if branch else None,
            # field NOT exposed in the schema
            # to be used while retrieving the node object
            "target_hash": branch.target if branch else None,
        }



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from swh.graphql.utils import utils

from .base_connection import BaseConnection, ConnectionData
from .base_node import BaseNode
from .origin import OriginNode


class BaseVisitNode(BaseNode):
    """
    Base resolver for all the visit nodes
    """

    @property
    def id(self):
        # FIXME, use a better id
        return utils.get_b64_string(f"{self.origin}-{str(self.visit)}")

    @property
    def visitId(self):  # To support the schema naming convention
        return self._node.visit


class OriginVisitNode(BaseVisitNode):
    """
    Node resolver for a visit requested directly with an origin URL
    and a visit ID
    """

    def _get_node_data(self):
        return self.archive.get_origin_visit(
            self.kwargs.get("originUrl"), int(self.kwargs.get("visitId"))
        )


class LatestVisitNode(BaseVisitNode):
    """
    Node resolver for the latest visit in an origin
    """

    _can_be_null = True
    obj: OriginNode

    def _get_node_data(self):
        # self.obj.url is the origin URL
        return self.archive.get_origin_latest_visit(
            origin_url=self.obj.url,
            visit_type=self.kwargs.get("visitType"),
            allowed_statuses=self.kwargs.get("allowedStatuses"),
            require_snapshot=self.kwargs.get("requireSnapshot"),
        )


class OriginVisitConnection(BaseConnection):
    """
    Connection resolver for the visit objects in an origin
    """

    obj: OriginNode

    _node_class = BaseVisitNode

    def _get_connection_data(self) -> ConnectionData:
        # self.obj.url is the origin URL
        return ConnectionData(
            paged_result=self.archive.get_origin_visits(
                self.obj.url,
                after=self._get_after_arg(),
                first=self._get_first_arg(),
                order=utils.get_storage_list_order(self.kwargs.get("sort")),
            )
        )



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from typing import Any, Dict, Optional

from aiocache import Cache
from ariadne.asgi import GraphQL
from starlette.applications import Starlette
from starlette.authentication import (
    AuthCredentials,
    AuthenticationBackend,
    UnauthenticatedUser,
)
from starlette.middleware import Middleware
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.middleware.cors import CORSMiddleware
from starlette.routing import Route

from swh.auth.starlette.backends import BearerTokenAuthBackend
from swh.core import config
from swh.search import get_search as get_swh_search
from swh.search.interface import SearchInterface
from swh.storage import get_storage as get_swh_storage
from swh.storage.interface import StorageInterface

from .client.view import explorer_page

graphql_cfg: Dict[str, Any] = {}
storage: Optional[StorageInterface] = None
search: Optional[SearchInterface] = None


def get_storage() -> StorageInterface:
    global storage
    if not storage:
        storage = get_swh_storage(**graphql_cfg["storage"])
    return storage


def get_search() -> SearchInterface:
    global search
    if not search:
        search = get_swh_search(**graphql_cfg["search"])
    return search


class AnonymousAuthBackend(AuthenticationBackend):
    async def authenticate(self, conn):
        return AuthCredentials(["anonymous"]), UnauthenticatedUser()


def load_and_check_config(config_path: Optional[str]) -> Dict[str, Any]:
    """Check the minimal configuration is set to run the api or raise an
       error explanation.

    Args:
        config_path: Path to the configuration file to load

    Raises:
        Error if the setup is not as expected

    Returns:
        configuration as a dict

    """
    if not config_path:
        raise EnvironmentError("Configuration file must be defined")

    if not os.path.exists(config_path):
        raise FileNotFoundError(f"Configuration file {config_path} does not exist")

    cfg = config.read(config_path)
    if "storage" not in cfg:
        raise KeyError("Missing 'storage' configuration")

    return cfg


def make_app_from_configfile():
    """Loading the configuration from a configuration file.

    SWH_CONFIG_FILENAME environment variable defines the
    configuration path to load.
    """
    from .app import schema, validation_rules
    from .errors import format_error, on_auth_error
    from .middlewares.logger import LogMiddleware

    global graphql_cfg

    if not graphql_cfg:
        config_path = os.environ.get("SWH_CONFIG_FILENAME")
        graphql_cfg = load_and_check_config(config_path)

    ariadne_app = GraphQL(
        schema,
        debug=graphql_cfg["debug"],
        introspection=graphql_cfg["introspection"],
        validation_rules=validation_rules,
        error_formatter=format_error,
    )

    if "auth" in graphql_cfg:
        auth_backend = BearerTokenAuthBackend(
            server_url=graphql_cfg["auth"]["server"],
            realm_name=graphql_cfg["auth"]["realm"],
            client_id=graphql_cfg["auth"]["client"],
            # FIXME, improve this with response cache implementation
            cache=Cache.from_url(url=graphql_cfg["auth"]["cache"]["url"]),
        )
    else:
        auth_backend = AnonymousAuthBackend()

    middleware = [
        Middleware(
            CORSMiddleware,
            # FIXME, restrict origins after deploying the JS client
            allow_origins=["*"],
            allow_methods=("GET", "POST", "OPTIONS"),
            allow_headers=["*"],
        ),
        Middleware(
            AuthenticationMiddleware,
            backend=auth_backend,
            on_error=on_auth_error,
        ),
        Middleware(LogMiddleware),
    ]

    # Mount under a starlette application
    application = Starlette(
        routes=[
            Route("/", ariadne_app, methods=["POST"], name="graphql_api"),
            Route("/", explorer_page, methods=["GET"]),
        ],
        middleware=middleware,
    )
    return application



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Optional

from . import utils
from ..data import get_origins


def get_origins_from_api(
    client, first: int, urlPattern: Optional[str] = None, **args
) -> tuple:
    query_str = """
    query getOrigins($first: Int!, $urlPattern: String) {
      origins(first: $first, urlPattern: $urlPattern) {
        nodes {
          url
        }
        totalCount
        pageInfo {
          hasNextPage
          endCursor
        }
      }
    }
    """
    return utils.get_query_response(
        client, query_str, first=first, urlPattern=urlPattern
    )


def test_get(client):
    data, _ = get_origins_from_api(client, first=10)
    assert len(data["origins"]["nodes"]) == len(get_origins())


def test_get_filter_by_pattern(client):
    data, _ = get_origins_from_api(client, first=10, urlPattern='"somewhere.org/den"')
    assert len(data["origins"]["nodes"]) == 1


def test_get_filter_by_non_existing_pattern(client):
    data, _ = get_origins_from_api(
        client, first=10, urlPattern='"somewhere.org/den/test/"'
    )
    assert len(data["origins"]["nodes"]) == 0


def test_basic_pagination(client):
    data, _ = get_origins_from_api(client, first=len(get_origins()))
    assert data["origins"]["totalCount"] is None
    assert len(data["origins"]["nodes"]) == len(get_origins())
    assert data["origins"]["pageInfo"] == {"hasNextPage": False, "endCursor": None}



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime

from swh.graphql.utils import utils


class TestUtils:
    def test_get_b64_string(self):
        assert utils.get_b64_string("testing") == "dGVzdGluZw=="

    def test_get_b64_string_binary(self):
        assert utils.get_b64_string(b"testing") == "dGVzdGluZw=="

    def test_get_encoded_cursor_is_none(self):
        assert utils.get_encoded_cursor(None) is None

    def test_get_encoded_cursor(self):
        assert utils.get_encoded_cursor(None) is None
        assert utils.get_encoded_cursor("testing") == "dGVzdGluZw=="

    def test_get_decoded_cursor_is_none(self):
        assert utils.get_decoded_cursor(None) is None

    def test_get_decoded_cursor(self):
        assert utils.get_decoded_cursor("dGVzdGluZw==") == "testing"

    def test_get_formatted_date(self):
        date = datetime.datetime(
            2015, 8, 4, 22, 26, 14, 804009, tzinfo=datetime.timezone.utc
        )
        assert utils.get_formatted_date(date) == "2015-08-04T22:26:14.804009+00:00"

    def test_get_local_paginated_data(self):
        source = [1, 2, 3, 4, 5]
        response = utils.get_local_paginated_data(source, first=50)
        assert response.paged_result.results == source
        assert response.paged_result.next_page_token is None

    def test_get_local_paginated_data_first_arg(self):
        source = [1, 2, 3, 4, 5]
        response = utils.get_local_paginated_data(source, first=2)
        assert response.paged_result.results == source[:2]
        assert response.paged_result.next_page_token == "2"

    def test_get_local_paginated_data_after_arg(self):
        source = [1, 2, 3, 4, 5]
        response = utils.get_local_paginated_data(source, first=2, after="2")
        assert response.paged_result.results == [3, 4]
        assert response.paged_result.next_page_token == "4"

        response = utils.get_local_paginated_data(source, first=2, after="3")
        assert response.paged_result.results == [4, 5]
        assert response.paged_result.next_page_token is None

    def test_get_local_paginated_data_endcursor_outside(self):
        source = [1, 2, 3, 4, 5]
        response = utils.get_local_paginated_data(source, first=2, after="10")
        assert response.paged_result.results == []
        assert response.paged_result.next_page_token is None



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Union

from ariadne import format_error as original_format_error
from graphql import GraphQLError, GraphQLSyntaxError
import sentry_sdk
from starlette.requests import Request
from starlette.responses import JSONResponse

from .errors import InvalidInputError, ObjectNotFoundError, PaginationError


def format_error(error: Union[GraphQLError, GraphQLSyntaxError], debug: bool = False):
    """
    Response error formatting
    """
    original_format = original_format_error(error, debug)
    if debug:
        # If debug is enabled, reuse Ariadne's formatting logic with stack trace
        return original_format
    # Errors raised by ariadne and graphql-core, it is safe to skip query syntax errors
    expected_base_errors = (GraphQLSyntaxError,)
    expected_errors = (ObjectNotFoundError, PaginationError, InvalidInputError)
    if not isinstance(error, expected_base_errors) and not isinstance(
        error.original_error, expected_errors
    ):
        # a crash, send to sentry
        sentry_sdk.capture_exception(error)
    formatted = error.formatted
    formatted["message"] = error.message
    # FIXME log the original_format to kibana (with stack trace)
    return formatted


def on_auth_error(request: Request, exc: Exception):
    # this error is raised outside the resolver context
    # using the default error formatter to log in sentry
    wrapped_error = GraphQLError("Authentication error", original_error=exc)
    return JSONResponse({"errors": [format_error(wrapped_error)]}, status_code=401)



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.graphql import resolvers
from swh.graphql.resolvers import resolvers as rs


class TestResolvers:
    """ """

    @pytest.fixture
    def dummy_node(self):
        return {"test": "test"}

    @pytest.mark.parametrize(
        "resolver_func, node_cls",
        [
            (rs.origin_resolver, resolvers.origin.OriginNode),
            (rs.visit_resolver, resolvers.visit.OriginVisitNode),
            (rs.latest_visit_resolver, resolvers.visit.LatestVisitNode),
            (
                rs.latest_visit_status_resolver,
                resolvers.visit_status.LatestVisitStatusNode,
            ),
            (rs.snapshot_resolver, resolvers.snapshot.SnapshotNode),
            (rs.revision_resolver, resolvers.revision.RevisionNode),
            (rs.revision_directory_resolver, resolvers.directory.RevisionDirectoryNode),
            (rs.release_resolver, resolvers.release.ReleaseNode),
            (rs.directory_resolver, resolvers.directory.DirectoryNode),
            (rs.content_by_hashes_resolver, resolvers.content.ContentbyHashesNode),
        ],
    )
    def test_node_resolver(self, mocker, dummy_node, resolver_func, node_cls):
        mock_get = mocker.patch.object(node_cls, "_get_node", return_value=dummy_node)
        node_obj = resolver_func(None, None)
        # assert the _get_node method is called on the right object
        assert isinstance(node_obj, node_cls)
        assert mock_get.assert_called

    @pytest.mark.parametrize(
        "resolver_func, connection_cls",
        [
            (rs.origins_resolver, resolvers.origin.OriginConnection),
            (rs.visits_resolver, resolvers.visit.OriginVisitConnection),
            (rs.origin_snapshots_resolver, resolvers.snapshot.OriginSnapshotConnection),
            (rs.visitstatus_resolver, resolvers.visit_status.VisitStatusConnection),
            (
                rs.snapshot_branches_resolver,
                resolvers.snapshot_branch.SnapshotBranchConnection,
            ),
            (rs.revision_parents_resolver, resolvers.revision.ParentRevisionConnection),
            (rs.revision_log_resolver, resolvers.revision.LogRevisionConnection),
            (
                rs.directory_entries_resolver,
                resolvers.directory_entry.DirectoryEntryConnection,
            ),
        ],
    )
    def test_connection_resolver(self, resolver_func, connection_cls):
        connection_obj = resolver_func(None, None)
        # assert the right object is returned
        assert isinstance(connection_obj, connection_cls)

    def test_union_resolver(self, mocker):
        obj = mocker.Mock()
        obj.is_type_of.return_value = "test"
        assert rs.union_resolver(obj) == "test"

    def test_binary_string_text_resolver(self):
        text = rs.binary_string_text_resolver(b"test", None)
        assert text == "test"

    def test_binary_string_base64_resolver(self):
        b64string = rs.binary_string_base64_resolver(b"test", None)
        assert b64string == "dGVzdA=="



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from . import utils
from ..data import get_snapshots


@pytest.mark.parametrize("snapshot", get_snapshots())
def test_get_snapshot(client, snapshot):
    query_str = """
    query getSnapshot($swhid: SWHID!) {
      snapshot(swhid: $swhid) {
        id
        swhid
        branches(first:5) {
          nodes {
            type
            name {
              text
            }
          }
        }
      }
    }
    """
    data, _ = utils.get_query_response(client, query_str, swhid=str(snapshot.swhid()))
    assert data["snapshot"]["swhid"] == str(snapshot.swhid())
    assert data["snapshot"]["id"] == snapshot.id.hex()
    assert len(data["snapshot"]["branches"]["nodes"]) == len(snapshot.branches)


def test_get_snapshot_missing_swhid(client):
    query_str = """
    query getSnapshot($swhid: SWHID!) {
      snapshot(swhid: $swhid) {
        swhid
      }
    }
    """
    utils.assert_missing_object(
        client,
        query_str,
        obj_type="snapshot",
        swhid="swh:1:snp:0949d7a8c96347dba09be8d79085b8207f345412",
    )


def test_get_snapshot_invalid_swhid(client):
    query_str = """
    query getSnapshot($swhid: SWHID!) {
      snapshot(swhid: $swhid) {
        swhid
      }
    }
    """
    errors = utils.get_error_response(client, query_str, swhid="swh:1:snp:invalid")
    assert len(errors) == 1
    assert "Input error: Invalid SWHID" in errors[0]["message"]



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


class ObjectNotFoundError(Exception):
    """ """

    msg: str = "Object error"

    def __init__(self, message, errors=None):
        super().__init__(f"{self.msg}: {message}")


class PaginationError(Exception):
    """ """

    msg: str = "Pagination error"

    def __init__(self, message, errors=None):
        super().__init__(f"{self.msg}: {message}")


class InvalidInputError(Exception):
    """ """

    msg: str = "Input error"

    def __init__(self, message, errors=None):
        super().__init__(f"{self.msg}: {message}")


class NullableObjectError(Exception):
    pass


class DataError(Exception):
    pass



# Copyright (C) 2022-2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from starlette.middleware.base import BaseHTTPMiddleware

from swh.core import statsd


class LogMiddleware(BaseHTTPMiddleware):
    # Starlette/ASGI middleware for logging request information

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.statsd = statsd.Statsd(namespace="swh_graphql")

    async def dispatch(self, request, call_next):
        if request.method == "POST":
            # This is a query request
            self.statsd.increment("query_total")
            # FIXME, add a metric for total query cost
            # FIXME, start a sentry transaction here
        with self.statsd.timed("query_seconds"):
            # total exec time metric (for both POST and GET)
            response = await call_next(request)
        return response



# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os

from starlette.templating import Jinja2Templates


async def explorer_page(request):
    from swh.graphql.server import graphql_cfg

    auth = graphql_cfg.get("auth")
    if auth and "public_server" not in auth:
        # ensure to not break already deployed service
        auth["public_server"] = auth["server"]

    templates = Jinja2Templates(directory=os.path.dirname(__file__))
    return templates.TemplateResponse(
        "explorer.html", {"request": request, "auth": auth}
    )



# Copyright (C) 2022-2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import List, Optional

from .base_connection import BaseList
from .content import BaseContentNode
from .directory import BaseDirectoryNode
from .release import BaseReleaseNode
from .revision import BaseRevisionNode
from .snapshot import BaseSnapshotNode


class ResolveSWHIDList(BaseList):
    def _get_results(self) -> Optional[List]:
        swhid = self.kwargs.get("swhid")
        object_type = swhid.object_type
        object_id = swhid.object_id
        nodes = None
        if object_type.name == "REVISION":
            self._node_class = BaseRevisionNode
            nodes = self.archive.get_revisions([object_id])
        elif object_type.name == "RELEASE":
            self._node_class = BaseReleaseNode
            nodes = self.archive.get_releases([object_id])
        elif object_type.name == "SNAPSHOT":
            self._node_class = BaseSnapshotNode
            # get_snapshot will return a single object
            nodes = [self.archive.get_snapshot(object_id, verify=True)]
        elif object_type.name == "DIRECTORY":
            self._node_class = BaseDirectoryNode
            # get_directory will return a single object
            nodes = [self.archive.get_directory(object_id, verify=True)]
        elif object_type.name == "CONTENT":
            self._node_class = BaseContentNode
            nodes = self.archive.get_contents(hashes={"sha1_git": object_id})
        if not nodes or nodes[0] is None:
            nodes = None
        return nodes



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from . import utils
from ..data import get_releases, get_snapshots_with_multiple_alias


def get_branches(client, **kwargs) -> tuple:
    query_str = """
    query getSnapshot($swhid: SWHID!, $first: Int!, $after: String, $types: [BranchType],
    $nameInclude: String, $excludePrefix: String) {
      snapshot(swhid: $swhid) {
        branches(first: $first, after: $after, types: $types, nameInclude: $nameInclude,
        nameExcludePrefix: $excludePrefix ) {
          pageInfo {
            endCursor
          }
          nodes {
            name {
              text
            }
            type
            target {
              type
              resolveChain {
                text
              }
              swhid
              node {
                __typename
                ...on Revision {
                  swhid
                }
                ...on Release {
                  swhid
                }
                ...on Content {
                  swhid
                }
                ...on Directory {
                  swhid
                }
                ...on Snapshot {
                  swhid
                }
              }
            }
          }
        }
      }
    }
    """
    return utils.get_query_response(client, query_str, **kwargs)


def test_get_data(client):
    swhid = "swh:1:snp:0e7f84ede9a254f2cd55649ad5240783f557e65f"
    data, errors = get_branches(client, swhid=swhid, first=10, types=["revision"])
    assert len(data["snapshot"]["branches"]["nodes"]) == 1
    # filter 'type' will return a single revision object and is used to assert data
    node = data["snapshot"]["branches"]["nodes"][0]
    assert node == {
        "name": {"text": "target/revision"},
        "type": "revision",
        "target": {
            "type": "revision",
            "resolveChain": None,
            "swhid": "swh:1:rev:66c7c1cd9673275037140f2abff7b7b11fc9439c",
            "node": {
                "__typename": "Revision",
                "swhid": "swh:1:rev:66c7c1cd9673275037140f2abff7b7b11fc9439c",
            },
        },
    }


def test_get_branches_with_one_level_alias(client):
    swhid = "swh:1:snp:0e7f84ede9a254f2cd55649ad5240783f557e65f"
    data, _ = get_branches(client, swhid=swhid, first=10, types=["alias"])
    node = data["snapshot"]["branches"]["nodes"][0]

    assert node == {
        "name": {"text": "target/alias"},  # original name
        "type": "alias",
        "target": {
            "swhid": "swh:1:rev:66c7c1cd9673275037140f2abff7b7b11fc9439c",
            "type": "revision",
            "resolveChain": [{"text": "target/revision"}],
            "node": {
                "__typename": "Revision",
                "swhid": "swh:1:rev:66c7c1cd9673275037140f2abff7b7b11fc9439c",
            },
        },
    }


def test_get_branches_alias_without_a_target(client):
    snapshot = get_snapshots_with_multiple_alias()[0]
    data, errors = get_branches(
        client,
        swhid=str(snapshot.swhid()),
        first=1,
        types=["alias"],
        nameInclude="alias1",
    )
    node = data["snapshot"]["branches"]["nodes"][0]
    assert node == {
        "name": {"text": "target/alias1"},  # original name
        "type": "alias",
        "target": {
            "node": None,
            "type": None,
            "swhid": None,
            "resolveChain": [{"text": "target/alias2"}, {"text": "target/alias3"}],
        },
    }


def test_get_branches_with_multiple_alias_redirects(client):
    snapshot = get_snapshots_with_multiple_alias()[1]
    data, errors = get_branches(
        client,
        swhid=str(snapshot.swhid()),
        first=1,
        types=["alias"],
        nameInclude="alias1",
    )
    node = data["snapshot"]["branches"]["nodes"][0]
    assert node == {
        "name": {"text": "target/alias1"},  # original name
        "type": "alias",
        "target": {
            "type": "release",
            "resolveChain": [{"text": "target/alias2"}, {"text": "target/release"}],
            "swhid": str(get_releases()[0].swhid()),
            "node": {"__typename": "Release", "swhid": str(get_releases()[0].swhid())},
        },
    }


def test_get_branches_with_too_many_alias_redirects(client):
    snapshot = get_snapshots_with_multiple_alias()[2]
    data, errors = get_branches(
        client,
        swhid=str(snapshot.swhid()),
        first=1,
        types=["alias"],
        nameInclude="alias1",
    )
    node = data["snapshot"]["branches"]["nodes"][0]
    assert node == {
        "name": {"text": "target/alias1"},  # original name
        "type": "alias",
        "target": {
            "node": None,
            "type": None,
            "swhid": None,
            "resolveChain": [
                {"text": "target/alias2"},
                {"text": "target/alias3"},
                {"text": "target/alias4"},
                {"text": "target/alias5"},
                {"text": "target/alias6"},
            ],
        },
    }


def test_get_branches_with_infinite_alias_redirects(client):
    snapshot = get_snapshots_with_multiple_alias()[3]
    data, errors = get_branches(
        client,
        swhid=str(snapshot.swhid()),
        first=1,
        types=["alias"],
        nameInclude="alias1",
    )
    node = data["snapshot"]["branches"]["nodes"][0]
    assert node == {
        "name": {"text": "target/alias1"},  # original name
        "type": "alias",
        "target": {
            "node": None,
            "type": None,
            "swhid": None,
            "resolveChain": [
                {"text": "target/alias2"},
                {"text": "target/alias1"},
                {"text": "target/alias2"},
            ],
        },
    }


@pytest.mark.parametrize(
    "filter_type, count, target_type, swhid_pattern",
    [
        ("revision", 1, "Revision", "swh:1:rev"),
        ("release", 1, "Release", "swh:1:rel"),
        ("directory", 1, "Directory", "swh:1:dir"),
        ("content", 0, "Content", "swh:1:cnt"),
        ("snapshot", 1, "Snapshot", "swh:1:snp"),
    ],
)
def test_get_type_filter(client, filter_type, count, target_type, swhid_pattern):
    swhid = "swh:1:snp:0e7f84ede9a254f2cd55649ad5240783f557e65f"
    data, _ = get_branches(client, swhid=swhid, first=10, types=[filter_type])
    assert len(data["snapshot"]["branches"]["nodes"]) == count
    for node in data["snapshot"]["branches"]["nodes"]:
        assert node["target"]["resolveChain"] is None
        assert node["target"]["swhid"].startswith(swhid_pattern)
        assert node["target"]["node"]["__typename"] == target_type


@pytest.mark.parametrize(
    "filter_types, count",
    [
        (["revision", "release"], 2),
        (["revision", "snapshot", "release"], 3),
    ],
)
def test_get_type_filter_multiple(client, filter_types, count):
    swhid = "swh:1:snp:0e7f84ede9a254f2cd55649ad5240783f557e65f"
    data, _ = get_branches(client, swhid=swhid, first=10, types=filter_types)
    assert len(data["snapshot"]["branches"]["nodes"]) == count


@pytest.mark.parametrize("name", ["rel", "rev", "non-exist"])
def test_get_name_include_filter(client, name):
    swhid = "swh:1:snp:0e7f84ede9a254f2cd55649ad5240783f557e65f"
    data, _ = get_branches(client, swhid=swhid, first=10, nameInclude=name)
    for node in data["snapshot"]["branches"]["nodes"]:
        assert name in node["name"]["text"]


@pytest.mark.parametrize("name", ["target", "target/dir"])
def test_get_name_exclude_prefix_filter(client, name):
    swhid = "swh:1:snp:0e7f84ede9a254f2cd55649ad5240783f557e65f"
    data, _ = get_branches(client, swhid=swhid, first=10, excludePrefix=name)
    for node in data["snapshot"]["branches"]["nodes"]:
        assert not node["name"]["text"].startswith(name)


@pytest.mark.parametrize("count", [1, 2])
def test_get_first_arg(client, count):
    swhid = "swh:1:snp:0e7f84ede9a254f2cd55649ad5240783f557e65f"
    data, _ = get_branches(client, swhid=swhid, first=count)
    assert len(data["snapshot"]["branches"]["nodes"]) == count


def test_get_after_arg(client):
    swhid = "swh:1:snp:0e7f84ede9a254f2cd55649ad5240783f557e65f"
    first_data, _ = get_branches(client, swhid=swhid, first=1)
    end_cursor = first_data["snapshot"]["branches"]["pageInfo"]["endCursor"]
    node_name = first_data["snapshot"]["branches"]["nodes"][0]["name"]["text"]
    second_data, _ = get_branches(client, swhid=swhid, first=3, after=end_cursor)
    branches = second_data["snapshot"]["branches"]
    assert len(branches["nodes"]) == 3
    for node in branches["nodes"]:
        assert node["name"]["text"] > node_name



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from typing import Any, Dict, Iterable, List, Optional, Tuple

from swh.graphql import server
from swh.model.model import (
    Content,
    Directory,
    DirectoryEntry,
    Origin,
    OriginVisit,
    OriginVisitStatus,
    Release,
    Revision,
    Sha1Git,
    Snapshot,
    SnapshotBranch,
)
from swh.model.swhids import ObjectType
from swh.storage.algos.origin import origin_get_latest_visit_status
from swh.storage.algos.snapshot import snapshot_resolve_branch_target
from swh.storage.interface import (
    HashDict,
    ListOrder,
    PagedResult,
    PartialBranches,
    StorageInterface,
)


class Archive:
    def __init__(self) -> None:
        self.storage: StorageInterface = server.get_storage()

    def get_origin(self, url: str) -> Optional[Origin]:
        return list(self.storage.origin_get(origins=[url]))[0]

    def get_origins(
        self, after: Optional[str] = None, first: int = 50
    ) -> PagedResult[Origin]:
        return self.storage.origin_list(page_token=after, limit=first)

    def get_origin_visits(
        self,
        origin_url: str,
        order: ListOrder,
        after: Optional[str] = None,
        first: int = 50,
    ) -> PagedResult[OriginVisit]:
        return self.storage.origin_visit_get(
            origin=origin_url, page_token=after, limit=first, order=order
        )

    def get_origin_visit(self, origin_url: str, visit_id: int) -> Optional[OriginVisit]:
        return self.storage.origin_visit_get_by(origin=origin_url, visit=visit_id)

    def get_origin_latest_visit(
        self,
        origin_url: str,
        visit_type: Optional[str] = None,
        allowed_statuses: Optional[List[str]] = None,
        require_snapshot: bool = False,
    ) -> Optional[OriginVisit]:
        return self.storage.origin_visit_get_latest(
            origin=origin_url,
            type=visit_type,
            allowed_statuses=allowed_statuses,
            require_snapshot=require_snapshot,
        )

    def get_visit_statuses(
        self,
        origin_url: str,
        visit_id: int,
        order: ListOrder,
        after: Optional[str] = None,
        first: int = 50,
    ) -> PagedResult[OriginVisitStatus]:
        return self.storage.origin_visit_status_get(
            origin=origin_url,
            visit=visit_id,
            page_token=after,
            limit=first,
            order=order,
        )

    def get_latest_visit_status(
        self,
        origin_url: str,
        visit_id: int,
        allowed_statuses: Optional[List[str]] = None,
        require_snapshot: bool = False,
    ) -> Optional[OriginVisitStatus]:
        return self.storage.origin_visit_status_get_latest(
            origin_url=origin_url,
            visit=visit_id,
            allowed_statuses=allowed_statuses,
            require_snapshot=require_snapshot,
        )

    def get_origin_snapshots(self, origin_url: str) -> List[Sha1Git]:
        return self.storage.origin_snapshot_get_all(origin_url=origin_url)

    def get_snapshot(
        self, snapshot_id: Sha1Git, verify: bool = True
    ) -> Optional[Snapshot]:
        # FIXME, change to accept list of snapshot_ids if needed
        if verify and not self.is_object_available(
            object_id=snapshot_id, object_type=ObjectType.SNAPSHOT
        ):
            # verify is True and the object is missing in the archive
            return None
        # Return a Snapshot model object; branches is initialized as empty
        # Same pattern is used in get_directory
        return Snapshot(id=snapshot_id, branches={})

    def get_snapshot_branches(
        self,
        snapshot: Sha1Git,
        after: bytes = b"",
        first: int = 50,
        target_types: Optional[List[str]] = None,
        name_include: Optional[bytes] = None,
        name_exclude_prefix: Optional[bytes] = None,
    ) -> Optional[PartialBranches]:
        return self.storage.snapshot_get_branches(
            snapshot_id=snapshot,
            branches_from=after,
            branches_count=first,
            target_types=target_types,
            branch_name_include_substring=name_include,
            branch_name_exclude_prefix=name_exclude_prefix,
        )

    def get_revisions(self, revision_ids: List[Sha1Git]) -> List[Optional[Revision]]:
        return self.storage.revision_get(revision_ids=revision_ids)

    def get_revision_log(
        self, revision_ids: List[Sha1Git], first: int = 50
    ) -> Iterable[Optional[Dict[str, Any]]]:
        return self.storage.revision_log(revisions=revision_ids, limit=first)

    def get_releases(self, release_ids: List[Sha1Git]) -> List[Optional[Release]]:
        return self.storage.release_get(releases=release_ids)

    def get_directory(
        self, directory_id: Sha1Git, verify: bool = True
    ) -> Optional[Directory]:
        # FIXME, change to accept list of directory_ids if needed
        if verify and not self.is_object_available(
            object_id=directory_id, object_type=ObjectType.DIRECTORY
        ):
            # verify is True and the object is missing in the archive
            return None
        # Return a Directory model object; entries is initialized as empty
        # Same pattern is used in get_snapshot
        return Directory(id=directory_id, entries=())

    def get_directory_entry_by_path(
        self, directory_id: Sha1Git, path: str
    ) -> Optional[Dict[str, Any]]:
        paths = [x.encode() for x in path.strip(os.path.sep).split(os.path.sep)]
        return self.storage.directory_entry_get_by_path(
            directory=directory_id, paths=paths
        )

    def get_directory_entries(
        self, directory_id: Sha1Git, after: Optional[bytes] = None, first: int = 50
    ) -> Optional[PagedResult[DirectoryEntry]]:
        return self.storage.directory_get_entries(
            directory_id=directory_id, limit=first, page_token=after
        )

    def is_object_available(self, object_id: bytes, object_type: ObjectType) -> bool:
        mapping = {
            ObjectType.CONTENT: self.storage.content_missing_per_sha1_git,
            ObjectType.DIRECTORY: self.storage.directory_missing,
            ObjectType.RELEASE: self.storage.release_missing,
            ObjectType.REVISION: self.storage.revision_missing,
            ObjectType.SNAPSHOT: self.storage.snapshot_missing,
        }
        return not list(mapping[object_type]([object_id]))

    def get_contents(self, hashes: HashDict) -> List[Content]:
        return self.storage.content_find(content=hashes)

    # def get_content_data(self, content_sha1: Sha1) -> Optional[bytes]:
    #     return self.storage.content_get_data(content=content_sha1)

    def get_branch_target(
        self,
        snapshot_id: Sha1Git,
        branch_obj: Optional[SnapshotBranch],
        max_length: int,
    ) -> Tuple[Optional[SnapshotBranch], List[bytes]]:
        return snapshot_resolve_branch_target(
            storage=self.storage,
            snapshot_id=snapshot_id,
            branch_obj=branch_obj,
            max_length=max_length,
        )

    def get_latest_origin_visit_status(
        self, origin: str, require_snapshot: bool = True
    ):
        return origin_get_latest_visit_status(
            storage=self.storage, origin_url=origin, require_snapshot=require_snapshot
        )



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Optional

from swh.graphql import server
from swh.search.interface import MinimalOriginDict, SearchInterface
from swh.storage.interface import PagedResult


class Search:
    def __init__(self):
        self.search: SearchInterface = server.get_search()

    def get_origins(
        self, query: str, after: Optional[str] = None, first: int = 50
    ) -> PagedResult[MinimalOriginDict]:
        return self.search.origin_search(
            url_pattern=query,
            page_token=after,
            limit=first,
        )



# Copyright (C) 2022 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from abc import ABC, abstractmethod
import binascii
from dataclasses import dataclass
from typing import Any, List, Optional, Type, Union

from graphql.type import GraphQLResolveInfo

from swh.graphql.backends.archive import Archive
from swh.graphql.backends.search import Search
from swh.graphql.errors import PaginationError
from swh.graphql.utils import utils
from swh.storage.interface import PagedResult

from .base_node import BaseNode


@dataclass
class PageInfo:
    hasNextPage: bool
    endCursor: Optional[str]


@dataclass
class ConnectionEdge:
    node: Any
    cursor: Optional[str]


@dataclass
class ConnectionData:
    paged_result: PagedResult
    total_count: Optional[int] = None


class BaseConnection(ABC):
    """
    Base resolver for all the connections
    """

    _node_class: Optional[Type[BaseNode]] = None
    _page_size: int = 50  # default page size (default value for the first arg)
    _max_page_size: int = 1000  # maximum page size(max value for the first arg)

    def __init__(self, obj, info, data=None, **kwargs):
        self.obj: Optional[BaseNode] = obj
        self.info: GraphQLResolveInfo = info
        self._connection_data: ConnectionData = data
        self.kwargs = kwargs
        # initialize commonly used vars
        self.archive = Archive()
        self.search = Search()

    @property
    def edges(self) -> List[ConnectionEdge]:
        """
        Return the list of connection edges, each with a cursor
        """
        return [
            ConnectionEdge(node=node, cursor=self._get_index_cursor(index, node))
            for (index, node) in enumerate(self.nodes)
        ]

    @property
    def nodes(self) -> List[Union[BaseNode, object]]:
        """
        Override if needed; return a list of objects

        If a node class is set, return a list of its (Node) instances
        else a list of raw results
        """
        if self._node_class is not None:
            return [
                self._node_class(
                    obj=self, info=self.info, node_data=result, **self.kwargs
                )
                for result in self.get_connection_data().paged_result.results
            ]
        return self.get_connection_data().paged_result.results

    @property
    def pageInfo(self) -> PageInfo:  # To support the schema naming convention
        # FIXME, add more details like startCursor
        return PageInfo(
            hasNextPage=bool(self.get_connection_data().paged_result.next_page_token),
            endCursor=utils.get_encoded_cursor(
                self.get_connection_data().paged_result.next_page_token
            ),
        )

    @property
    def totalCount(self) -> Optional[int]:  # To support the schema naming convention
        """
        Will be None for most of the connections
        override if needed/possible
        """
        return self.get_connection_data().total_count

    def get_connection_data(self) -> ConnectionData:
        """
        Cache to avoid multiple calls to the backend :meth:`_get_connection_data`
        """
        if self._connection_data is None:
            # FIXME, make this call async (not for v1)
            self._connection_data = self._get_connection_data()
        return self._connection_data

    @abstractmethod
    def _get_connection_data(self) -> ConnectionData:
        """
        Override to fetch data
        """
        # FIXME, make this call async (not for v1)

    def _get_after_arg(self):
        """
        Return the decoded next page token. Override to support a different
        cursor type
        """
        # different implementation is used in SnapshotBranchConnection
        try:
            cursor = utils.get_decoded_cursor(self.kwargs.get("after"))
        except (UnicodeDecodeError, binascii.Error) as e:
            raise PaginationError("Invalid value for argument 'after'", errors=e)
        return cursor

    def _get_first_arg(self) -> int:
        """ """
        # page_size is set to 50 by default
        # Input type check is not required; It is defined in schema as an int
        first = self.kwargs.get("first", self._page_size)
        if first < 0 or first > self._max_page_size:
            raise PaginationError(
                f"Value for argument 'first' is invalid; it must be between 0 and {self._max_page_size}"  # noqa: B950
            )
        return first

    def _get_index_cursor(self, index: int, node: Any) -> Optional[str]:
        """
        Get the cursor to the given item index
        """
        # default implementation which works with swh-storage pagaination
        # override this function to support other types (eg: SnapshotBranchConnection)
        offset_index = self._get_after_arg() or "0"
        index_cursor = int(offset_index) + index
        return utils.get_encoded_cursor(str(index_cursor))


class BaseList(ABC):
    """
    Base class to be used for simple lists that do not require
    pagination; eg resolveSWHID entrypoint
    """

    _node_class: Optional[Type[BaseNode]] = None

    def __init__(self, obj, info, results=None, **kwargs):
        self.obj: Optional[BaseNode] = obj
        self.info: GraphQLResolveInfo = info
        self.kwargs = kwargs
        self._results: List = results

        self.archive = Archive()

    def get_results(self) -> Optional[List[Any]]:
        if self._results is None:
            # To avoid multiple calls to the backend
            self._results = self._get_results()

        if self._node_class is not None:
            # convert list items to node objects
            return [
                self._node_class(
                    obj=self.obj, info=self.info, node_data=result, **self.kwargs
                )
                for result in self._results
            ]
        return self._results

    @abstractmethod
    def _get_results(self) -> Optional[List]:
        """
        Override for desired behaviour
        return a list of objects
        """
